import yaml
import os
import copy

def convert_general_dict(vib_dict,*,halfway_basis=True):
    new_vib_dict = copy.deepcopy(vib_dict)
    d = vib_dict['displacement']
    if type(d) is list:
        flag_3LS = True
        d2 = d[1]
        d = d[0]
    else:
        flag_3LS = False
    
    if 'alpha' in vib_dict.keys() and 'potential' in vib_dict.keys():
        raise Exception('Cannot specify potential and alpha simultaneously')
    try:
        if flag_3LS:
            p0,p1,p2 = vib_dict['potential']
            alpha2 = (p2[0]/p0[0])**(1/2)
        else:
            p0,p1 = vib_dict['potential']
        
        alpha = (p1[0]/p0[0])**(1/2)
    except KeyError:
        try:
            if flag_3LS:
                alpha, alpha2 = vib_dict['alpha']
            else:
                alpha = vib_dict['alpha']
            new_vib_dict.pop('alpha')
        except KeyError:
            alpha = 1
            alpha2 = 1
        p0 = [1]
        p1 = [alpha**2]
        p2 = [alpha2**2]
        
    try:
        if flag_3LS:
            k0,k1, k2 = vib_dict['kinetic']
        else:
            k0,k1 = vib_dict['kinetic']
    except KeyError:
        k0 = [1]
        k1 = [1]
        if flag_3LS:
            k2 = [1]
    
    new_d = d*alpha**(1/4)/2
    if flag_3LS:
        new_d2 = (d2-d/2)*alpha**(1/4)
    
    try:
        if flag_3LS:
            l, l2 = vib_dict['reorganization']
        else:
            l = vib_dict['reorganization']
    except KeyError:
        if flag_3LS:
            l = -alpha**2 * d**2
            l2 = -alpha2**2 * d2**2
        else:
            l = -alpha**2 * d**2

    if halfway_basis:
        k0_modifier = [alpha**(i/4) for i in range(2,len(k0)+2)]
        k1_modifier = [alpha**(i/4) for i in range(2,len(k1)+2)]
        p0_modifier = [alpha**(-i/4) for i in range(2,len(p0)+2)]
        p1_modifier = [alpha**(-i/4) for i in range(2,len(p1)+2)]
        if flag_3LS:
            k2_modifier = [alpha**(i/4) for i in range(2,len(k2)+2)]
            p2_modifier = [alpha**(-i/4) for i in range(2,len(p2)+2)]
    else:
        k0_modifier = [1 for i in range(2,len(k0)+2)]
        k1_modifier = [1 for i in range(2,len(k1)+2)]
        p0_modifier = [1 for i in range(2,len(p0)+2)]
        p1_modifier = [1 for i in range(2,len(p1)+2)]
        if flag_3LS:
            k2_modifier = [1 for i in range(2,len(k2)+2)]
            p2_modifier = [1 for i in range(2,len(p2)+2)]
        
    new_k0 = [a*b for a,b in zip(k0,k0_modifier)]
    new_k1 = [a*b for a,b in zip(k1,k1_modifier)]
    
    new_p0 = [a*b for a,b in zip(p0,p0_modifier)]
    new_p1 = [a*b for a,b in zip(p1,p1_modifier)]
    if flag_3LS:
        new_k2 =  [a*b for a,b in zip(k2,k2_modifier)]
        new_p2 =  [a*b for a,b in zip(p2,p2_modifier)]

    new_vib_dict['reorganization'] = [0,l]
    new_vib_dict['kinetic'] = [new_k0,new_k1]
    new_vib_dict['potential'] = [new_p0,new_p1]

    if halfway_basis:
        new_vib_dict['displacement'] = [-new_d,new_d]
    else:
        new_vib_dict['displacement'] = [0,d]

    if flag_3LS:
        new_vib_dict['reorganization'].append(l2)
        new_vib_dict['kinetic'].append(new_k2)
        new_vib_dict['potential'].append(new_p2)
        if halfway_basis:
            new_vib_dict['displacement'].append(new_d2)
        else:
            new_vib_dict['displacement'].append(d2)
        
    return new_vib_dict

def convert(base_path,*,halfway_basis=True):
    simple_file_name = os.path.join(base_path,'simple_params.yaml')
    parameter_file_name = os.path.join(base_path,'params.yaml')
    with open(simple_file_name,'r') as yamlstream:
        simple = yaml.load(yamlstream,Loader=yaml.SafeLoader)
    
    params = dict()
    try:
        params['RWA'] = simple['RWA']
    except KeyError:
        params['RWA'] = True
    params['dipoles'] = simple['dipoles']
    params['site_energies'] = simple['site_energies']
    params['site_couplings'] = simple['site_couplings']
    params['initial truncation size'] = simple['truncation_size']
    try:
        params['number eigenvalues'] = simple['num_eigenvalues']
        params['eigenvalue precision'] = simple['eigenvalue_precision']
    except KeyError:
        pass
    
    try:
        params['maximum_manifold'] = simple['maximum_manifold']
    except KeyError:
        pass

    try:
        params['bath'] = simple['bath']
    except KeyError:
        pass

    simple_vibrations = simple['vibrations']

    vibrations = [convert_general_dict(vib,halfway_basis=halfway_basis) for vib in simple_vibrations]
        
    params['vibrations'] = vibrations

    with open(parameter_file_name,'w+') as yamlstream:
        yamlstream.write('### This file was generated by UFSS, do not modify unless you know what you are doing \n')
        yaml.dump(params,yamlstream)

    



"""
This doesn't work unless there are four pulses. Thus for things like TA, one 
must divide both the pump and probe pulses in two, so that there are 4 effective
pulses. The two copies of the pump should be multiplied by a factor of 2, while 
the two copies of the probe pulse should
"""

#Standard python libraries
import time
import itertools

#Dependencies
import numpy as np

"""The following definitions of kdelvec and kdelvec6 are 
based upon the formulas given in Appendix B of Molecular Quantum 
Electrodynamics, by Akbar Salam

I4_mat and I6_mat are created using the formulae provided in 
"On three‐dimensional rotational averages"
by Andrews, D.L. and Thirunamachandran, T., doi:10.1063/1.434725

Unittest is available to check I4_mat and I6_mat against the matrices that were
explicitly written out in Salam's book.
"""

def kdel(x,y):
    """Kronecker Delta"""
    if x == y:
        return 1
    else:
        return 0

def kdel2(a,b,c,d):
    """Product of 2 Kronecker Deltas"""
    return kdel(a,b)*kdel(c,d)

def kdel3(a,b,c,d,e,f):
    """Product of 2 Kronecker Deltas"""
    return kdel(a,b)*kdel(c,d)*kdel(e,f)

def kdelvec4(i,j,k,l):
    """Length 3 vector of Kronecker Delta products, as defined in """
    vec = [kdel2(i,j,k,l),
           kdel2(i,k,j,l),
           kdel2(i,l,j,k)]
    return np.array(vec)

def kdelvec6(i,j,k,l,m,n):
    """Length 3 vector of Kronecker Delta products, as defined in """
    vec = [kdel3(i,j,k,l,m,n),
           kdel3(i,j,k,m,n,l),
           kdel3(i,j,k,n,l,m),
           kdel3(i,k,j,l,m,n),
           kdel3(i,k,j,m,n,l),
           kdel3(i,k,j,n,l,m),
           kdel3(i,l,j,k,m,n),
           kdel3(i,l,j,m,k,n),
           kdel3(i,l,j,n,k,m),
           kdel3(i,m,j,k,n,l),
           kdel3(i,m,j,l,k,n),
           kdel3(i,m,j,n,k,l),
           kdel3(i,n,j,k,l,m),
           kdel3(i,n,j,l,k,m),
           kdel3(i,n,j,m,k,l)]
    return np.array(vec)

def kdelvec(indices):
    if len(indices) == 4:
        return kdelvec4(*indices)
    elif len(indices) == 6:
        return kdelvec6(*indices)
    else:
        raise Exception('Isotropic averaging is only implemented for rank 4 and rank 6 tensors')

S4 = np.zeros((3,3))
for i in range(3):
    for j in range(3):
        for k in range(3):
            for l in range(3):
                vec1 = kdelvec4(i,j,k,l)
                vec2 = kdelvec4(i,j,k,l)
                S4 += np.outer(vec1,vec2)
I4_mat = np.linalg.inv(S4)

S6 = np.zeros((15,15))
for i in range(3):
    for j in range(3):
        for k in range(3):
            for l in range(3):
                for m in range(3):
                    for n in range(3):
                        vec1 = kdelvec6(i,j,k,l,m,n)
                        vec2 = kdelvec6(i,j,k,l,m,n)
                        S6 += np.outer(vec1,vec2)
I6_mat = np.linalg.inv(S6)

def get_In_mat(indices):
    if len(indices) == 4:
        return I4_mat
    elif len(indices) == 6:
        return I6_mat
    else:
        raise Exception('Isotropic averaging is only implemented for rank 4 and rank 6 tensors')

class IsotropicAverage(object):
    """This class performs the isotropic average of the 4th order tensor
        which is the material response produced by 4-wave mixing process"""

    def __init__(self,spectra_calculator,*,diagrams='all'):
        """Takes as input a ufss object that calculates 4-wave mixing spectra,
        and calculates the isotropically averaged signal, given a lab-frame polarization"""
        self.sc = spectra_calculator
        self.diagrams = diagrams
        self.signal_dict = {}

    def molecular_frame_signal(self,mol_polarization):
        self.sc.set_polarization_sequence(mol_polarization)
        if self.diagrams == 'all':
            signal = self.sc.calculate_signal_all_delays(composite_diagrams=False)
        else:
            signal = self.sc.calculate_diagrams_all_delays(self.diagrams)
        return signal
    
    def get_polarization_options(self):
        """Determines which polarizations are necessary, given the cartesian 
        structure of the dipole operator"""
        xyz = ['x','y','z']

        pol_options = []
        for i in range(3):
            # Check to see if the dipole operator has any non-zero components along the given
            # molecular frame axis, if the dipole exists only in the x-y plane, for example,
            # then we can avoid doing quite a few unnecessary calculations!
            try:
                random_mu_key = list(self.sc.mu.keys())[0]
                test_mu = self.sc.mu[random_mu_key]
            except AttributeError:
                random_mu_key = list(self.sc.H_mu.keys())[0]
                test_mu = self.sc.H_mu[random_mu_key]
            if type(test_mu) is np.ndarray:
                if not np.allclose(test_mu[:,:,i],0):
                    pol_options.append(xyz[i])
            elif type(test_mu) is list:
                if not np.allclose(test_mu[i][:,:],0):
                    pol_options.append(xyz[i])

        return pol_options
    
    def get_signal_key(self,polarization_sequence):
        return ''.join(polarization_sequence)

    def averaged_signal(self,lab_polarizations,*,return_signal=False,reset=False):
        t0 = time.time()
        self.pol = lab_polarizations
        if reset:
            self.signal_dict = {}

        left_vec = kdelvec(self.pol)

        num_pol = len(self.pol)

        pol_options = self.get_polarization_options()

        pol_options_N = [pol_options]*num_pol

        pol_options_iter = itertools.product(*pol_options_N)

        In_mat = get_In_mat(self.pol)

        new_signal = True
        for pol in pol_options_iter:
            # generate the vector of kronecker delta products
            right_vec = kdelvec(pol)
            if np.allclose(right_vec,0):
                # If the vector is 0, don't bother!
                pass
            else:
                # If not, set the polarization sequence, do the calculation, and
                # add the weight given by the isotropic weight matrix, In_mat
                # Note the the polarization sequences are not the lab frame
                # polarization sequence of the pulses.
                key = self.get_signal_key(pol)
                try:
                    mol_frame_signal = self.signal_dict[key]
                except KeyError:
                    mol_frame_signal = self.molecular_frame_signal(pol)
                    self.signal_dict[key] = mol_frame_signal
                weight = In_mat.dot(right_vec)
                weight = np.dot(left_vec,weight)
                if new_signal:
                    signal = weight * mol_frame_signal
                    new_signal = False
                else:
                    signal += weight * mol_frame_signal

        self.signal = signal

        self.calculation_time = time.time() - t0
        if return_signal:
            return signal

    def save(self,save_name,pulse_delay_names=[],*,use_base_path=True):
        if save_name == 'auto':
            save_name = 'IsotropicAverage_'+''.join(self.pol)
        self.sc.signal = self.signal
        self.sc.calculation_time = self.calculation_time
        self.sc.save(save_name,pulse_delay_names,use_base_path=use_base_path)

class FWMIsotropicAverage(IsotropicAverage):
    """For backwards compatibility"""

    def __init__(self,spectra_calculator,lab_polarization,*,diagrams='all'):
        """Takes as input a ufss object that calculates 4-wave mixing spectra,
        and calculates the isotropically averaged signal, given a lab-frame polarization"""
        self.sc = spectra_calculator
        self.pol = lab_polarization
        self.diagrams = diagrams
        self.signal_dict = {}

    def averaged_signal(self,*,return_signal=False):
        sig = super().averaged_signal(self.pol,return_signal=return_signal)
        if return_signal:
            return sig

    def save(self,save_name,pulse_delay_names=[],*,use_base_path=True):
        if save_name == 'auto':
            save_name = 'FWM_IsotropicAverage_'+''.join(self.pol)
        super().save(save_name,pulse_delay_names=pulse_delay_names,
                     use_bath_path=use_base_path)   


import unittest
import numpy as np
import os
import yaml

import ufss

def run_HLG():
    cm_to_eV = 1/8065.54429    # cm^-1 to eV ; easier to just spell cm
    ev_to_THz = 1/4.1357E-15/1E12  # self-explanatory


    e_b = 13500*cm_to_eV

    e_a = 14700*cm_to_eV

    # transition coupling energies

    J_AB = -420*cm_to_eV  # one exciton coupling

    # transition dipole magnitudes

    # 0 -> 1 transition
    mu_a = np.array([1.2,0,0])

    mu_b = np.array([1,0,0])


    # bath parameters

    nu_a = 1.0

    nu_b = 1.66

    # v_f_average = (v_af + v_bf)/2

    lam = 180*cm_to_eV    # lower-case lambda
    gamma = 300*cm_to_eV  # capital lambda

    S = 0.18              # S, the Huang-Rhys factor
    Omega = 1280*cm_to_eV # Bath oscillation frequency
    Gamma = 10*cm_to_eV   # damping lower-case gamma


    # relaxation parameters are in fs or ps ; 
    # for the purposes of the HLG, we convert them to rates, in fs^-1, and then to eV radians

    # the 1->0 relaxation is dependent upon the polymer length, so is defined in the next cell

    def make_2LS_dimer():
        num_pairs = 1
        params_folder = os.path.join('fixtures','2LS_dimer_b')
        os.makedirs(params_folder,exist_ok=True)

        ### Making closed system parameters ###
        num_sites = 2*num_pairs # total number of sites

        site_energies = []

        for i in range(num_pairs):
            # site energies e_a and f_a for SQA monomer
            energies_a = e_a
            site_energies.append(energies_a)
            # site energies e_b and f_b for SQB monomer
            energies_b = e_b
            site_energies.append(energies_b)

        J_list = []

        for i in range(num_sites):
            for j in range(i+1,num_sites):
                # there's only coupling of adjacent units ; if i and j differ by more than 1 (i.e. if for J_i,j we have j-i>1,)
                if j-i > 1:
                    J = 0
                else : 
                    J = J_AB
                J_list.append(J)

        site_couplings = [J_list]

        dipoles = np.zeros((num_sites,3))
        # identical dipoles for all sites
        dipoles[0,:] = mu_a
        dipoles[1,:] = mu_b

        for i in range(2,num_sites-1,2):
            # dipole operator for SQA unit
            dipoles[i,...] = dipoles[0,...]

            # dipole operator for SQB unit
            dipoles[i+1,...] = dipoles[1,...]

        dipoles_list = dipoles.tolist()

        ### open system parameters ###

        # e -> g relaxation rate
        tau_r = (1710.2 - 69.1*num_pairs) # in ps

        k_r = 1/tau_r  # in ps^-1

        k_r = k_r/ ev_to_THz / (2*np.pi) # in eV radians ; no need for the *1000 factor since k_r is in ps^-1

        kT = 0.025  # assuming room temperature

        # site-bath coupling of singly-excited state, A-type monomer
        site_bath_a = {'cutoff_frequency':gamma,
                    'coupling': lam,
                    'temperature':kT,
                    'cutoff_function':'brownian',
                    'S': S,
                    'Gamma' : Gamma,          
                    'Omega' : Omega,
                    'nu': nu_a,
                    'spectrum_type':'ohmic'}

        # site-bath coupling of singly-excited states, B-type monomer
        site_bath_b = {'cutoff_frequency':gamma,
                    'coupling': lam,
                    'temperature':kT,
                    'cutoff_function':'brownian',
                    'S': S,
                    'Gamma' : Gamma,          
                    'Omega' : Omega,
                    'nu': nu_b,
                    'spectrum_type':'ohmic'}

        # relaxation bath for 1->0 processes
        relax_bath = {'dephasing_rate':0.01, # maybe does nothing!
                    'relaxation_rate':k_r,
                    'temperature':kT,
                        'spectrum_type':'white-noise'}

        Redfield_bath = {'secular':True,
                'site_bath':[site_bath_a,site_bath_b],
                'site_bath_option' : [0,1]*num_pairs,
                'vibration_bath':site_bath_a,
                'site_internal_conversion_bath':relax_bath}

        params = {
            'site_energies':site_energies,
            'site_couplings':site_couplings,
            'dipoles':dipoles_list,
            'bath':Redfield_bath,
            'truncation_size':1,
            'vibrations':[]}

        with open(os.path.join(params_folder,'simple_params.yaml'),'w+') as new_file:
            yaml.dump(params,new_file)

        ufss.HLG.run(params_folder,conserve_memory=False)

    make_2LS_dimer()

def calculate_PPP(order,*,include_wtau_DC_terms=True,include_linear=True):
    cm_to_eV = 1/8065.54429    # cm^-1 to eV ; easier to just spell cm
    e_b = 13500*cm_to_eV
    center = e_b #center on squaraine B's transition energy
    open_folder = os.path.join('fixtures','2LS_dimer_b','open')
    td = ufss.signals.PumpPumpProbe(open_folder,conserve_memory=False,
                                    include_linear=include_linear,
                                    include_wtau_DC_terms=include_wtau_DC_terms)

    #impulsive or finite pulses
    td.set_impulsive_pulses(center)
    td.engine.efields = [np.array([0.5]),np.array([0.5]),np.array([1])]
    #td.set_identical_gaussians(1,center,M=101,delta=10)

    # highest deisred order in perturbation theory
    td.set_highest_order(order)

    # set the time-grid, 'auto' only works for open systems, otherwise
    # specify an exponential decay rate as a float imposed on the decay of P(t)
    td.set_t(0.05,dt=3)

    # inter-pulse delay times
    # coherence times
    tau = np.array([0])
    # populations times
    T = np.array([1E-2])
    td.set_pulse_delays([tau,T])
    td.calculate_signal_all_delays()
    return td

def calculate_TA(order,include_linear=True):
    cm_to_eV = 1/8065.54429   # cm^-1 to eV ; easier to just spell cm
    e_b = 13500*cm_to_eV
    center = e_b #center on squaraine B's transition energy
    open_folder = os.path.join('fixtures','2LS_dimer_b','open')
    ta = ufss.signals.TransientAbsorption(open_folder,conserve_memory=False,
                                          include_linear=include_linear)

    #impulsive or finite pulses
    ta.set_impulsive_pulses(center)
    #ta.set_identical_gaussians(1,center,M=101,delta=10)

    # highest deisred order in perturbation theory
    ta.set_highest_order(order)

    # set the time-grid, 'auto' only works for open systems, otherwise
    # specify an exponential decay rate as a float imposed on the decay of P(t)
    ta.set_t(0.05,dt=3)

    # populations times
    T = np.array([1E-2])
    ta.set_pulse_delays([T])
    ta.calculate_signal_all_delays()
    return ta

class test_TA_vs_PPP(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        run_HLG()

    def test_1(self):
        order = 25
        ppp = calculate_PPP(order)
        ta = calculate_TA(order)
        for i in range((order+1)//2):
            y1 = ppp.get_signal_order(2*i+1,tau_dft=False)[0,0,:]
            y2 = ta.get_signal_order(2*i+1)[0,:]
            res = np.allclose(y1,y2,atol=1E-20)
            with self.subTest(res=res):
                self.assertTrue(res)

    def test_2(self):
        order = 25
        ppp = calculate_PPP(order,include_linear=False,include_wtau_DC_terms=False)
        ta = calculate_TA(order,include_linear=False)
        for i in range(1,(order+1)//2):
            y1 = ppp.get_signal_order(2*i+1,tau_dft=False)[0,0,:]
            y2 = ta.get_signal_order(2*i+1)[0,:]
            res = np.allclose(y1,y2,atol=1E-20)
            with self.subTest(res=res):
                self.assertFalse(res)

if __name__ == '__main__':
    unittest.main()


import numpy as np
import ufss
import os
from ufss.signals import SpectroscopyBase

class TransientAbsorption(SpectroscopyBase):
    def __init__(self,file_name,*,engine_name = 'UF2',include_linear=False,
                 conserve_memory=False,detection_type='polarization'):
        super().__init__(file_name,engine_name=engine_name,
                         conserve_memory=conserve_memory,
                         detection_type=detection_type)

        
        if include_linear:
            self.pdc = ((0,0),(1,0))
        else:
            self.pdc = ((1,1),(1,0))
        self.save_name = 'transient_absorption_signals'
        
    def set_hightest_order(self,highest_order):
        self.engine.include_higher_orders((highest_order-1,1))

    def get_signal(self,lam,*,max_order=np.inf):
        """lam is the perturbative parameter
        Args:
            lam (float) : unit-less electric field amplitude scaling factor
"""
        lam_list = [lam,1]
        signal = super().get_signal(lam_list,max_order=max_order)

        return signal

    def set_pulse_delays(self,pulse_delays):
        super().set_pulse_delays(pulse_delays)
        self.T = pulse_delays[0]
    
    def plot_signal(self,lam):
        signal = self.get_signal(lam)
        wt = self.engine.w
        T = self.engine.all_pulse_delays[0]
        ufss.signals.plot2D(T,wt,signal,part='real')

    def plot_order(self,order,T_index):
        signal = self.get_signal_order(order)
        wt = self.engine.w
        T = self.engine.all_pulse_delays[0]
        ufss.signals.plot2D(T,wt,signal,part='real')



import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse import save_npz, csr_matrix
import warnings
import os

class ManualL:
    def __init__(self,L,mu_ket_up,*,output='uf2',savedir=''):
        self.L = L
        self.mu_ket_up = mu_ket_up
        self.output = output
        if savedir=='':
            savedir = os.getcwd()
        self.base_path = os.path.join(savedir,'open')
        print(self.base_path)
        os.makedirs(self.base_path,exist_ok=True)
        if output == 'uf2':
            self.save_L(self.base_path)
            self.eigfun(self.L)
            self.save_eigensystem(self.base_path)
            if len(mu_ket_up.shape) == 2:
                self.save_RWA_mu(self.base_path)
            elif len(mu_ket_up.shape) == 3:
                self.save_RWA_mu3D(self.base_path)
        elif output == 'RKE':
            print(self.base_path)
            self.save_L(self.base_path)
            self.save_RWA_mu_site_basis(self.base_path)

    def save_L(self,dirname):
        print(os.path.join(dirname,'L.npz'))
        save_npz(os.path.join(dirname,'L.npz'),csr_matrix(self.L))
            

    def eigfun(self,L,*,check_eigenvectors = True,invert = True,populations_only = False):
        eigvals, eigvecs = np.linalg.eig(L)

        eigvals = np.round(eigvals,12)
        sort_indices = eigvals.argsort()
        eigvals.sort()
        eigvecs = eigvecs[:,sort_indices]
        for i in range(eigvals.size):
            max_index = np.argmax(np.abs(eigvecs[:,i]))
            if np.real(eigvecs[max_index,i]) < 0:
                eigvecs[:,i] *= -1
            if eigvals[i] == 0:
                # eigenvalues of 0 correspond to thermal distributions,
                # which should have unit trace in the Hamiltonian space
                if populations_only:
                    trace_norm = eigvecs[:,i].sum()
                    eigvecs[:,i] = eigvecs[:,i] / trace_norm
                else:
                    shape = int(np.sqrt(eigvals.size))
                    trace_norm = eigvecs[:,i].reshape(shape,shape).trace()
                    eigvecs[:,i] = eigvecs[:,i] / trace_norm

        if invert:
            eigvecs_left = np.linalg.pinv(eigvecs)
        else:
            eigvals_left, eigvecs_left = np.linalg.eig(L.T)

            eigvals_left = np.round(eigvals_left,12)
            sort_indices_left = eigvals_left.argsort()
            eigvals_left.sort()
            eigvecs_left = eigvecs_left[:,sort_indices_left]
            eigvecs_left = eigvecs_left.T
            for i in range(eigvals_left.size):
                    norm = np.dot(eigvecs_left[i,:],eigvecs[:,i])
                    eigvecs_left[i,:] *= 1/norm

        if check_eigenvectors:
            LV = L.dot(eigvecs)
            D = eigvecs_left.dot(LV)
            if np.allclose(D,np.diag(eigvals),rtol=1E-10,atol=1E-10):
                pass
            else:
                warnings.warn('Using eigenvectors to diagonalize Liouvillian does not result in the expected diagonal matrix to tolerance, largest deviation is {}'.format(np.max(np.abs(D - np.diag(eigvals)))))

        self.eigenvalues = eigvals
        self.eigenvectors = {'left':eigvecs_left,'right':eigvecs}

        return eigvals, eigvecs, eigvecs_left

    def save_eigensystem(self,dirname):
        np.savez(os.path.join(dirname,'right_eigenvectors.npz'),all_manifolds = self.eigenvectors['right'])
        np.savez(os.path.join(dirname,'left_eigenvectors.npz'),all_manifolds = self.eigenvectors['left'])
        np.savez(os.path.join(dirname,'eigenvalues.npz'),all_manifolds = self.eigenvalues)

    def mu3D_eigentransform(self,mu):
        evl = self.eigenvectors['left']
        ev = self.eigenvectors['right']
        
        mu_t = np.zeros(mu.shape,dtype='complex')
        for i in range(3):
            mu_t[:,:,i] = np.dot(np.dot(evl,mu[:,:,i]),ev)
        return mu_t

    def mask_mu3D(self,mu):
        mu_mask_tol = 10
        
        mu_mask = np.zeros(mu.shape[:2],dtype='bool')
        mu_abs = np.sqrt(np.sum(np.abs(mu)**2,axis=2))
        mu_mask[:,:] = np.round(mu_abs,mu_mask_tol)[:,:]
        mu_masked = mu * mu_mask[:,:,np.newaxis]

        return mu_mask, mu_masked

    def save_RWA_mu3D(self,dirname,*,mask=True):

        H_size = self.mu_ket_up.shape[0]
        mu_dtype= self.mu_ket_up.dtype
        L_size = H_size**2
        II = np.eye(H_size)
        mu_ket_up = np.zeros((L_size,L_size,3),dtype=mu_dtype)
        mu_ket_down = np.zeros((L_size,L_size,3),dtype=mu_dtype)
        mu_bra_up = np.zeros((L_size,L_size,3),dtype=mu_dtype)
        mu_bra_down = np.zeros((L_size,L_size,3),dtype=mu_dtype)
        
        for i in range(3):
            mu_ket_up[:,:,i] = np.kron(self.mu_ket_up[:,:,i],II.T)
            mu_ket_down[:,:,i] = np.kron(np.conjugate(self.mu_ket_up[:,:,i].T),II.T)
            mu_bra_up[:,:,i] = np.kron(II,np.conjugate(self.mu_ket_up[:,:,i]))
            mu_bra_down[:,:,i] = np.kron(II,self.mu_ket_up[:,:,i].T)

        mu_ket_up_t = self.mu3D_eigentransform(mu_ket_up)
        mu_ket_down_t = self.mu3D_eigentransform(mu_ket_down)
        mu_bra_up_t = self.mu3D_eigentransform(mu_bra_up)
        mu_bra_down_t = self.mu3D_eigentransform(mu_bra_down)

        np.savez(os.path.join(dirname,'mu.npz'),ket_up=mu_ket_up_t,bra_up=mu_bra_up_t,
                 ket_down=mu_ket_down_t,bra_down=mu_bra_down_t)

        if mask:
            ket_up_t_mask, mu_ket_up_t_masked = self.mask_mu3D(mu_ket_up_t)
            ket_down_t_mask, mu_ket_down_t_masked = self.mask_mu3D(mu_ket_down_t)
            bra_up_t_mask, mu_bra_up_t_masked = self.mask_mu3D(mu_bra_up_t)
            bra_down_t_mask, mu_bra_down_t_masked = self.mask_mu3D(mu_bra_down_t)

            np.savez(os.path.join(dirname,'mu_boolean.npz'),ket_up=ket_up_t_mask,bra_up=bra_up_t_mask,
                     ket_down=ket_down_t_mask,bra_down=bra_down_t_mask)
            np.savez(os.path.join(dirname,'mu_pruned.npz'),ket_up=mu_ket_up_t_masked,
                     bra_up=mu_bra_up_t_masked,ket_down=mu_ket_down_t_masked,
                     bra_down=mu_bra_down_t_masked)

    def save_RWA_mu(self,dirname,*,mask=True):
        evl = self.eigenvectors['left']
        ev = self.eigenvectors['right']
        
        II = np.eye(self.mu_ket_up.shape[0])

        mu_ket_up = np.kron(self.mu_ket_up,II.T)
        mu_ket_down = np.kron(np.conjugate(self.mu_ket_up.T),II.T)
        mu_bra_up = np.kron(II,np.conjugate(self.mu_ket_up))
        mu_bra_down = np.kron(II,self.mu_ket_up.T)

        mu_mask_tol = 10
        
        mu_ket_up_t = np.dot(np.dot(evl,mu_ket_up),ev)
        mu_ket_up_3d = np.zeros((mu_ket_up_t.shape[0],mu_ket_up_t.shape[0],3),dtype='complex')
        mu_ket_up_3d[:,:,0] = mu_ket_up_t

        mu_bra_up_t = np.dot(np.dot(evl,mu_bra_up),ev)
        mu_bra_up_3d = np.zeros((mu_bra_up_t.shape[0],mu_bra_up_t.shape[0],3),dtype='complex')
        mu_bra_up_3d[:,:,0] = mu_bra_up_t

        mu_ket_down_t = np.dot(np.dot(evl,mu_ket_down),ev)
        mu_ket_down_3d = np.zeros((mu_ket_down_t.shape[0],mu_ket_down_t.shape[0],3),dtype='complex')
        mu_ket_down_3d[:,:,0] = mu_ket_down_t

        mu_bra_down_t = np.dot(np.dot(evl,mu_bra_down),ev)
        mu_bra_down_3d = np.zeros((mu_bra_down_t.shape[0],mu_bra_down_t.shape[0],3),dtype='complex')
        mu_bra_down_3d[:,:,0] = mu_bra_down_t

        if mask:
            ket_up_mask = np.zeros(mu_ket_up_t.shape,dtype='bool')
            ket_up_mask[:,:] = np.round(mu_ket_up_t,mu_mask_tol)[:,:]
            mu_ket_up_t_masked = mu_ket_up_t * ket_up_mask
            mu_ket_up_3d_masked = np.zeros((mu_ket_up_t.shape[0],mu_ket_up_t.shape[0],3),dtype='complex')
            mu_ket_up_3d_masked[:,:,0] = mu_ket_up_t_masked

            bra_up_mask = np.zeros(mu_bra_up_t.shape,dtype='bool')
            bra_up_mask[:,:] = np.round(mu_bra_up_t,mu_mask_tol)[:,:]
            mu_bra_up_t_masked = mu_bra_up_t * bra_up_mask
            mu_bra_up_3d_masked = np.zeros((mu_ket_up_t.shape[0],mu_ket_up_t.shape[0],3),dtype='complex')
            mu_bra_up_3d_masked[:,:,0] = mu_bra_up_t_masked

            ket_down_mask = np.zeros(mu_ket_down_t.shape,dtype='bool')
            ket_down_mask[:,:] = np.round(mu_ket_down_t,mu_mask_tol)[:,:]
            mu_ket_down_t_masked = mu_ket_down_t * ket_down_mask
            mu_ket_down_3d_masked = np.zeros((mu_ket_down_t.shape[0],mu_ket_down_t.shape[0],3),dtype='complex')
            mu_ket_down_3d_masked[:,:,0] = mu_ket_down_t_masked

            bra_down_mask = np.zeros(mu_bra_down_t.shape,dtype='bool')
            bra_down_mask[:,:] = np.round(mu_bra_down_t,mu_mask_tol)[:,:]
            mu_bra_down_t_masked = mu_bra_down_t * bra_down_mask
            mu_bra_down_3d_masked = np.zeros((mu_ket_down_t.shape[0],mu_ket_down_t.shape[0],3),dtype='complex')
            mu_bra_down_3d_masked[:,:,0] = mu_bra_down_t_masked

            np.savez(os.path.join(dirname,'mu.npz'),ket_up=mu_ket_up_3d,bra_up=mu_bra_up_3d,
                     ket_down=mu_ket_down_3d,bra_down=mu_bra_down_3d)
            # np.savez(os.path.join(dirname,'eigenvalues.npz'),all_manifolds=self.eigenvalues)
            # np.savez(os.path.join(dirname,'right_eigenvectors.npz'),all_manifolds=ev)
            # np.savez(os.path.join(dirname,'left_eigenvectors.npz'),all_manifolds=evl)
            np.savez(os.path.join(dirname,'mu_boolean.npz'),ket_up=ket_up_mask,bra_up=bra_up_mask,
                     ket_down=ket_down_mask,bra_down=bra_down_mask)
            np.savez(os.path.join(dirname,'mu_pruned.npz'),ket_up=mu_ket_up_3d_masked,
                     bra_up=mu_bra_up_3d_masked,ket_down=mu_ket_down_3d_masked,
                     bra_down=mu_bra_down_3d_masked)

        else:
            np.savez(os.path.join(dirname,'mu.npz'),ket_up=mu_ket_up_3d,bra_up=mu_bra_up_3d,
                     ket_down=mu_ket_down_3d,bra_down=mu_bra_down_3d)
            # np.savez(os.path.join(dirname,'eigenvalues.npz'),all_manifolds=self.eigenvalues)
            # np.savez(os.path.join(dirname,'right_eigenvectors.npz'),all_manifolds=ev)
            # np.savez(os.path.join(dirname,'left_eigenvectors.npz'),all_manifolds=evl)

    def save_RWA_mu_site_basis(self,dirname):
        
        II = np.eye(self.mu_ket_up.shape[0])
        mu_ket_up = np.kron(self.mu_ket_up,II.T)
        mu_ket_down = np.kron(self.mu_ket_up.T,II.T)
        mu_bra_up = np.kron(II,self.mu_ket_up)
        mu_bra_down = np.kron(II,self.mu_ket_up.T)

        mu_mask_tol = 10
        
        mu_ket_up_3d = np.zeros((mu_ket_up.shape[0],mu_ket_up.shape[0],3),dtype='complex')
        mu_ket_up_3d[:,:,0] = mu_ket_up

        mu_bra_up_3d = np.zeros((mu_bra_up.shape[0],mu_bra_up.shape[0],3),dtype='complex')
        mu_bra_up_3d[:,:,0] = mu_bra_up

        mu_ket_down_3d = np.zeros((mu_ket_down.shape[0],mu_ket_down.shape[0],3),dtype='complex')
        mu_ket_down_3d[:,:,0] = mu_ket_down

        mu_bra_down_3d = np.zeros((mu_bra_down.shape[0],mu_bra_down.shape[0],3),dtype='complex')
        mu_bra_down_3d[:,:,0] = mu_bra_down

        np.savez(os.path.join(dirname,'mu_site_basis.npz'),ket_up=mu_ket_up_3d,bra_up=mu_bra_up_3d,
                     ket_down=mu_ket_down_3d,bra_down=mu_bra_down_3d)



#Standard python libraries
import os
import warnings
import time

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
from scipy.sparse import csr_matrix

from ufss.perturbative_calculations import UF2BaseClass
from ufss.perturbative_calculations import perturbative_container
from ufss.perturbative_calculations import OpenBaseClass

class UF2OpenEngine(OpenBaseClass,UF2BaseClass):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the eigenvalues of the unperturbed 
        hamiltonian and the material dipole operator evaluated in the
        eigenbasis of the unperturbed hamiltonian.

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        num_conv_points (int): number of desired points for linear 
            convolution. Also number of points used to resolve all optical
            pulse shapes
        dt (float): time spacing used to resolve the shape of all optical
            pulses
        initial_state (int): index of initial state for psi^0

"""
    def __init__(self,file_path,*,detection_type = 'polarization',
                 conserve_memory=False):
        UF2BaseClass.__init__(self,file_path,detection_type=detection_type)
        OpenBaseClass.__init__(self,detection_type = detection_type)
        self.slicing_time = 0
        self.interpolation_time = 0
        self.expectation_time = 0
        self.next_order_expectation_time = 0
        self.convolution_time = 0
        self.extend_time = 0
        self.mask_time = 0
        self.dipole_time = 0
        self.automation_time = 0
        self.diagram_to_signal_time = 0
        self.diagram_generation_counter = 0
        self.number_of_diagrams_calculated = 0
        self.efield_mask_time = 0
        self.dipole_down_dot_product_time = 0
        self.reshape_and_sum_time = 0
        self.rho_to_polarization_time = 0
        self.next_order_time = 0

        self.next_order_counter = 0

        self.check_for_zero_calculation = False
        self.interaction_picture_shift = True
        self.interaction_picture_calculations = True

        self.sparsity_threshold = .1

        self.conserve_memory = conserve_memory

        self.load_eigensystem()

        self.set_rho_shapes()

        self.load_H_eigensystem()

        try:
            self.load_H_mu()
            self.H_mu_flag = True
        except FileNotFoundError:
            self.load_mu()
            self.H_mu_flag = False
            warnings.warn('Could not find H_mu, defaulting to L_mu for calculating expectation values, which is slower')

        if not self.conserve_memory:
            self.load_mu()

        if self.detection_type == 'fluorescence':
            self.load_fl_yield()

        self.set_maximum_manifold()
        

    ### Methods for calculating composite diagrams

    def add_impulsive_rhos(self,ra,rb):
        """Add two density matrix objects together
"""
        manifold_key = ra.manifold_key
        pdc = ra.pdc
        sim_a = ra.simultaneous
        sim_b = rb.simultaneous
        if sim_a != sim_b:
            raise Exception('Cannot add impulsive rhos if they do not have the same number of simultaneous interactions')
        else:
            simultaneous = sim_a
        if np.allclose(ra.t0,rb.t0):
            t = ra.t[1:2]
            rho_a = ra.f[:,1]
            rho_b = rb.f[:,1]
            pulse_number = ra.pulse_number
            t0 = ra.t0
        else:
            raise Exception('Cannot add two impulsive rhos unless they have the same value of t0')

        rho = np.zeros((ra.bool_mask.size,t.size),dtype='complex')
        rho[ra.bool_mask,:] = rho_a[:,np.newaxis]
        rho[rb.bool_mask,:] += rho_b[:,np.newaxis]
        rho *= 2 * simultaneous

        bool_mask = np.logical_or(ra.bool_mask,rb.bool_mask)
        rho = rho[bool_mask,:]
        
        rab=perturbative_container(t,rho,bool_mask,pulse_number,manifold_key,pdc,t0,
                          simultaneous = simultaneous)
        return rab
        

    def add_rhos(self,ra,rb):
        """Add two density matrix objects together
"""
        if not np.allclose(ra.pdc,rb.pdc):
            raise Exception('Cannot add density matrices with different phase-discrimination conditions')
        if 'all_manifolds' in self.manifolds:
            pass
        else:
            if ra.manifold_key != rb.manifold_key:
                raise Exception('Cannot add perturbative_container objects that exist in different manifolds')
            
        if ra.impulsive:
            return self.add_impulsive_rhos(ra,rb)
        
        manifold_key = ra.manifold_key
        
        pdc = ra.pdc
        
        tmin = min(ra.t[0],rb.t[0])
        tmax = max(ra.t[-1],rb.t[-1])
        dt = min(ra.t[1]-ra.t[0],rb.t[1]-rb.t[0])
        t = np.arange(tmin,tmax+dt*0.9,dt)
        rho = np.zeros((ra.bool_mask.size,t.size),dtype='complex')
        rho_a = ra(t)
        rho_b = rb(t)

        if ra.t0 > rb.t0:
            t0 = ra.t0
            pulse_number = ra.pulse_number
        else:
            t0 = rb.t0
            pulse_number = rb.pulse_number

        try:
            eva = self.eigenvalues['all_manifolds'][ra.bool_mask]
            evb = self.eigenvalues['all_manifolds'][rb.bool_mask]
        except KeyError:
            eva = self.eigenvalues[ra.manifold_key][ra.bool_mask]
            evb = self.eigenvalues[rb.manifold_key][rb.bool_mask]

        if self.interaction_picture_shift:
            t_exp_a = t0 - ra.t0
            t_exp_b = t0 - rb.t0
            rho_a *= np.exp(eva[:,np.newaxis]*t_exp_a)
            rho_b *= np.exp(evb[:,np.newaxis]*t_exp_b)
        else:
            pass

        rho[ra.bool_mask,:] = rho_a
        rho[rb.bool_mask,:] += rho_b

        bool_mask = np.logical_or(ra.bool_mask,rb.bool_mask)
        rho = rho[bool_mask,:]
        
        rab=perturbative_container(t,rho,bool_mask,pulse_number,manifold_key,pdc,t0)
        return rab

    def set_efields(self,times_list,efields_list,centers_list,
                    phase_discrimination,*,reset_calculations = True,
                    plot_fields = False):

        max_calculation_times = [np.max(np.abs(t)) for t in times_list]
        max_calc_time = np.max(max_calculation_times)
        largest_decay_rates = []
        for key in self.eigenvalues.keys():
            gamma_max = -np.min(np.real(self.eigenvalues[key]))
            largest_decay_rates.append(gamma_max)
        largest_decay_rate = np.max(largest_decay_rates)
        if largest_decay_rate * max_calc_time > 40:
            self.interaction_picture_calculations = False
        else:
            self.interaction_picture_calculations = True
        
        UF2BaseClass.set_efields(self,times_list,efields_list,centers_list,
                                 phase_discrimination,
                                 reset_calculations = reset_calculations,
                                 plot_fields = plot_fields)
        
        # Initialize unperturbed density matrix
        self.set_rho0()

    def convert_manifold_keys(self):
        old_manifolds = self.manifolds
        new_manifolds = []
        for k,b in self.manifolds:
            new_key = ''.join([k,',',b])
            new_manifolds.append(new_key)
        for old_key,new_key in zip(old_manifolds,new_manifolds):
            self.eigenvalues[new_key] = self.eigenvalues.pop(old_key)
            self.eigenvectors[new_key] = self.eigenvectors.pop(old_key)
            if self.conserve_memory:
                self.left_eigenvectors[new_key] = self.left_eigenvectors.pop(old_key)

        self.manifolds = new_manifolds

    ########## specific to UF2

    def save_timing(self):
        save_dict = {'UF2_calculation_time':self.calculation_time}
        np.savez(os.path.join(self.base_path,'UF2_calculation_time.npz'),**save_dict)

    def get_rho(self,t,rho_obj,*,reshape=True,original_L_basis=True):
        manifold_key = rho_obj.manifold_key
        mask = rho_obj.bool_mask
        if self.interaction_picture_shift:
            t_exp = t - rho_obj.t0
        else:
            t_exp = t
        
        try:
            e = self.eigenvalues['all_manifolds'][mask]
            ev = self.eigenvectors['all_manifolds'][:,mask]
            ket_size, bra_size = self.rho_shapes['all_manifolds']
        except KeyError:
            e = self.eigenvalues[manifold_key][mask]
            ev = self.eigenvectors[manifold_key][:,mask]
            ket_size, bra_size = self.rho_shapes[manifold_key]

        rho = rho_obj(t)*np.exp(e[:,np.newaxis]*t_exp[np.newaxis,:])
        if original_L_basis:
            new_rho = ev.dot(rho)
        else:
            new_rho = rho
        if reshape:
            if original_L_basis:
                new_rho = new_rho.reshape(ket_size,bra_size,rho.shape[-1])
            else:
                raise Exception('Cannot reshape without moving to original L basis')
        return new_rho

    def set_rho0(self):
        """Creates the unperturbed density matarix by finding the 0 
            eigenvalue of the ground-state manifold, which should correspond
            to a thermal distribution
"""
        try:
            ev = self.eigenvalues['all_manifolds']
        except KeyError:
            ev = self.eigenvalues['0,0']

        t = self.efield_times[0] # can be anything of the correct length
        
        initial_state = np.where(np.isclose(ev,0,atol=1E-12))[0]
        rho0 = np.ones((1,t.size),dtype=complex)
        bool_mask = np.zeros(ev.size,dtype='bool')
        bool_mask[initial_state] = True

        if bool_mask.sum() != 1:
            warnings.warn('Could not automatically determine the initial thermal state. User must specify the initial condition, rho^(0), manually')
            return None

        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)

        t0 = 0

        self.rho0 = perturbative_container(t,rho0,bool_mask,None,'0,0',pdc,t0,
                                  interp_kind='zero',interp_left_fill=1)

    def set_rho0_manual_L_eigenbasis(self,manifold_key,bool_mask,weights):
        """
"""
        ev = self.eigenvalues[manifold_key][bool_mask]

        t = self.efield_times[0] # can be anything of the correct length
        
        rho0 = np.ones((bool_mask.sum(),t.size),dtype=complex) * weights[:,np.newaxis]

        if manifold_key == 'all_manifolds':
            manifold_key = '0,0'

        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)

        t0 = 0

        self.rho0 = perturbative_container(t,rho0,bool_mask,None,manifold_key,pdc,t0,
                                  interp_kind='zero',interp_left_fill=1)

    def set_rho0_manual(self,rho0,*,manifold_key = 'all_manifolds'):
        """Set the initial condition.  Must be done after setting the pulse shapes
        Args:
            rho0 (2D np.array) : the initial density matrix, in the basis that
                the Liouvillian was defined in
            manfifold_key (str) : manifold in which initial density matrix is 
                defined (usually 'all_manifolds' (default) or '00')
"""
        try:
            evl = self.left_eigenvectors[manifold_key]
        except AttributeError:
            self.load_left_eigenvectors()
            evl = self.left_eigenvectors[manifold_key]

        rho0_flat = rho0.flatten() # ufss always works with vectors
        

        rho0_eig = evl.dot(rho0_flat) # transform into Liouvillian eigenbasis
        
        nonzero_inds = np.where(np.abs(rho0_eig) > 1E-12)[0]

        bool_mask = np.zeros(rho0_eig.size,dtype='bool')
        bool_mask[nonzero_inds] = True

        rho0_trimmed = rho0_eig[nonzero_inds]

        t = self.efield_times[0]
        time_dependence = np.ones(t.size)
        rho0 = rho0_trimmed[:,np.newaxis] * time_dependence[np.newaxis,:]
        
        if manifold_key == 'all_manifolds':
            manifold_key = '0,0'

        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)

        t0 = 0

        self.rho0 = perturbative_container(t,rho0,bool_mask,None,manifold_key,pdc,t0,
                                  interp_kind='zero',interp_left_fill=1)

    def load_eigensystem(self):
        """Load in known eigenvalues and eigenvectors. Must be stored as a 
            numpy archive file
"""
        eigval_save_name = os.path.join(self.base_path,'eigenvalues.npz')
        eigvec_save_name = os.path.join(self.base_path,'right_eigenvectors.npz')
        with np.load(eigval_save_name) as eigval_archive:
            self.manifolds = list(eigval_archive.keys())
            self.eigenvalues = {key:eigval_archive[key] for key in self.manifolds}
        with np.load(eigvec_save_name,allow_pickle=True) as eigvec_archive:
            self.eigenvectors = dict()
            for key in self.manifolds:
                ev = eigvec_archive[key]
                if ev.dtype == np.dtype('O'):
                    self.eigenvectors[key] = ev[()]
                elif self.check_sparsity(ev):
                    self.eigenvectors[key] = csr_matrix(ev)
                else:
                    self.eigenvectors[key] = ev

        if self.conserve_memory:
            self.load_left_eigenvectors()

        if len(self.manifolds[0]) == 2:
            self.convert_manifold_keys()

    def load_left_eigenvectors(self):
        left_eigvec_save_name = os.path.join(self.base_path,'left_eigenvectors.npz')
        with np.load(left_eigvec_save_name,allow_pickle=True) as left_eigvec_archive:
            self.left_eigenvectors = dict()
            for key in self.manifolds:
                evl = left_eigvec_archive[key]
                if evl.dtype == np.dtype('O'):
                    self.left_eigenvectors[key] = evl[()]
                elif self.check_sparsity(evl):
                    self.left_eigenvectors[key] = csr_matrix(evl)
                else:
                    self.left_eigenvectors[key] = evl
                        
    def set_rho_shapes(self):
        self.rho_shapes = dict()
        if 'all_manifolds' in self.manifolds:
            L_size = self.eigenvalues['all_manifolds'].size
            H_size = int(np.sqrt(L_size))
            self.rho_shapes['all_manifolds'] = (H_size,H_size)
        else:
            H_sizes = dict()
            for key in self.manifolds:
                ket_key, bra_key = key.split(',')
                if ket_key == bra_key:
                    L_size = self.eigenvalues[key].size
                    H_size = int(np.sqrt(L_size))
                    H_sizes[ket_key] = H_size
            for key in self.manifolds:
                ket_key, bra_key = key.split(',')
                ket_size = H_sizes[ket_key]
                bra_size = H_sizes[bra_key]
                self.rho_shapes[key] = (ket_size,bra_size)
            

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
            be stored as a .npz file, and must contain at least one array, each with three 
            indices: (new manifold eigenfunction, old manifold eigenfunction, 
            cartesian coordinate)."""
        file_name = os.path.join(self.base_path,'mu.npz')
        file_name_pruned = os.path.join(self.base_path,'mu_pruned.npz')
        file_name_bool = os.path.join(self.base_path,'mu_boolean.npz')
        try:
            mu_boolean_archive = np.load(file_name_bool)
            with np.load(file_name_bool) as mu_boolean_archive:
                self.mu_boolean = {key:mu_boolean_archive[key] for key in mu_boolean_archive.keys()}
            file_name = file_name_pruned
            pruned = True
        except FileNotFoundError:
            pruned = False
        with np.load(file_name) as mu_archive:
            self.mu = {key:mu_archive[key] for key in mu_archive.keys()}
        if pruned == False:
            self.mu_boolean = dict()
            for key in self.mu.keys():
                self.mu_boolean[key] = np.ones(self.mu[key].shape[:2],dtype='bool')
        sparse_flags = []
        for key in self.mu.keys():
            mu_2D = np.sum(np.abs(self.mu[key])**2,axis=-1)
            sparse_flags.append(self.check_sparsity(mu_2D))
        sparse_flags = np.array(sparse_flags)
        if np.allclose(sparse_flags,True):
            self.sparse_mu_flag = True
        else:
            self.sparse_mu_flag = False

        for key in self.mu.keys():
            mu_x = self.mu[key][...,0]
            mu_y = self.mu[key][...,1]
            mu_z = self.mu[key][...,2]

            if self.sparse_mu_flag:
                self.mu[key] = [csr_matrix(mu_x),csr_matrix(mu_y),csr_matrix(mu_z)]
            else:
                self.mu[key] = [mu_x,mu_y,mu_z]

        if 'all_manifolds' in self.manifolds:
            pass
        elif ',' in key:
            pass
        else:
            self.convert_mu_keys()

    ### Tools for recursively calculating perturbed wavepackets using TDPT

    def dipole_matrix(self,pulse_number,key,ket_flag=True,up_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector,
            if ket_flag = False then uses the bra-interaction"""
        overlap_matrix = OpenBaseClass.dipole_matrix(self,pulse_number,key,
                                                     ket_flag=ket_flag,
                                                     up_flag=up_flag)
        try:
            boolean_matrix = self.mu_boolean[key]
        except KeyError:
            if ket_flag:
                key = 'ket'
            else:
                key = 'bra'
            if up_flag:
                key += '_up'
            else:
                key += '_down'
            boolean_matrix = self.mu_boolean[key]

        return boolean_matrix, overlap_matrix

    def electric_field_mask(self,pulse_number,key,conjugate_flag=False):
        """This method determines which molecular transitions will be 
supported by the electric field.  We assume that the electric field has
0 amplitude outside the minimum and maximum frequency immplied by the 
choice of dt and num_conv_points.  Otherwise we will inadvertently 
alias transitions onto nonzero electric field amplitudes.
"""
        ta = time.time()

        try:
            mask = self.efield_masks[pulse_number][key]
        except KeyError:
            starting_key, ending_key = key.split('_to_')
            efield_t = self.efield_times[pulse_number]
            efield_w = self.efield_frequencies[pulse_number]
            if conjugate_flag:
                center = -self.centers[pulse_number]
            else:
                center = self.centers[pulse_number]
            try:
                eig_starting = self.eigenvalues['all_manifolds']
                eig_ending = self.eigenvalues['all_manifolds']
            except KeyError:
                eig_starting = self.eigenvalues[starting_key]
                eig_ending = self.eigenvalues[ending_key]
            # imag part corresponds to the energetic transitions
            diff = np.imag(eig_ending[:,np.newaxis] - eig_starting[np.newaxis,:])



            if efield_t.size == 1:
                mask = np.ones(diff.shape,dtype='bool')
            else:
                # The only transitions allowed by the electric field shape are
                inds_allowed = np.where((diff + center > efield_w[0]) & (diff + center < efield_w[-1]))
                mask = np.zeros(diff.shape,dtype='bool')
                mask[inds_allowed] = 1
            
            self.efield_masks[pulse_number][key] = mask

            tb = time.time()
            self.efield_mask_time += tb - ta

        return mask

    def mask_dipole_matrix(self,boolean_matrix,overlap_matrix,
                           starting_manifold_mask,*,next_manifold_mask = None):
        """Takes as input the boolean_matrix and the overlap matrix that it 
            corresponds to. Also requires the starting manifold mask, which specifies
            which states have non-zero amplitude, given the signal tolerance requested.
            Trims off unnecessary starting elements, and ending elements. If 
            next_manifold_mask is None, then the masking is done automatically
            based upon which overlap elements are nonzero. If next_manifold_mask is
            a 1D numpy boolean array, it is used as the mask for next manifold."""
        t0 = time.time()
        if np.all(starting_manifold_mask == True):
            pass
        else:
            boolean_matrix = boolean_matrix[:,starting_manifold_mask]
            overlap_matrix = overlap_matrix[:,starting_manifold_mask]

        #Determine the nonzero elements of the new psi, in the
        #eigenenergy basis, n_nonzero
        if type(next_manifold_mask) is np.ndarray:
            n_nonzero = next_manifold_mask
        else:
            n_nonzero = np.any(boolean_matrix,axis=1)
        if np.all(n_nonzero == True):
            pass
        else:
            overlap_matrix = overlap_matrix[n_nonzero,:]

        t1 = time.time()
        self.mask_time += t1-t0

        return overlap_matrix, n_nonzero

    def get_H_mu(self,pulse_number,key,rotating_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector"""
        overlap_matrix = OpenBaseClass.get_H_mu(self,pulse_number,key,
                                                 rotating_flag=rotating_flag)

        boolean_matrix = np.ones(overlap_matrix.shape,dtype='bool')

        return boolean_matrix, overlap_matrix

    def rho_matrix_to_L_vector(self,rho,manifold_key,*,
                               diagonal_L_basis=True):
        t_size = rho.shape[-1]
        rho_vec_size = rho.shape[0]*rho.shape[1]
        rho = rho.reshape(rho_vec_size,t_size)
        if diagonal_L_basis:
            if 'all_manifolds' in self.manifolds:
                evl = self.left_eigenvectors['all_manifolds']
            else:
                evl = self.left_eigenvectors[manifold_key]
                
            rho = evl.dot(rho)

        return rho

    def rho_L_vector_to_matrix(self,rho,manifold_key,bool_mask):
        try:
            ev = self.eigenvectors['all_manifolds'][:,bool_mask]
            ket_size,bra_size = self.rho_shapes['all_manifolds']
        except KeyError:
            ev = self.eigenvectors[manifold_key][:,bool_mask]
            ket_size,bra_size = self.rho_shapes[manifold_key]

        rho = ev.dot(rho)
        tsize = rho.shape[-1]
        
        rho = rho.reshape(ket_size,bra_size,tsize)

        return rho

    def mu_dot_rho(self,t,rho_in,pulse_number,ket_flag=True,up_flag=True,
                   next_manifold_mask = None,mu_type='L'):
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        boolean_mu, mu = self.get_mu(pulse_number,old_manifold_key,
                                     new_manifold_key,ket_flag=ket_flag,
                                     up_flag=up_flag,mu_type=mu_type)
            
        if mu_type=='H':
            rho =self.get_rho(t,rho_in,reshape=True,original_L_basis=True)

            if ket_flag:
                rho = np.einsum('ij,jkl',mu,rho)
            else:
                rho = np.einsum('ijl,jk',rho,mu)
            
            rho = self.rho_matrix_to_L_vector(rho,new_manifold_key)
            
            n_nonzero = np.ones(rho.shape[0],dtype='bool')
            zero_inds = np.where(np.isclose(rho[:,-1],0,atol=1E-160))
            n_nonzero[zero_inds] = False
            rho = rho[n_nonzero,:]
            
        elif mu_type=='L':
            rho =self.get_rho(t,rho_in,reshape=False,original_L_basis=False)
            m_nonzero = rho_in.bool_mask

            # e_mask = self.electric_field_mask(pulse_number,mu_key,conjugate_flag=conjugate_flag)
            # boolean_matrix = boolean_matrix * e_mask
            # overlap_matrix = overlap_matrix * e_mask

            mu,n_nonzero=self.mask_dipole_matrix(boolean_mu,mu,m_nonzero,
                                    next_manifold_mask = next_manifold_mask)

            rho = mu.dot(rho)
        else:
            raise Exception("mu_type must be 'H' or 'L'")
            
        return rho, n_nonzero

    def mu_exp_val(self,t,rho_in,pulse_number,ket_flag=True,up_flag=False,
                   mu_type='H'):
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        if not self.H_mu_flag:
            if mu_type == 'H':
                warnings.warn('Using L_mu for expectation values because H_mu is not available')
                mu_type = 'L'

        
        boolean_mu, mu = self.get_mu(pulse_number,old_manifold_key,
                                     new_manifold_key,ket_flag=ket_flag,
                                     up_flag=up_flag,mu_type=mu_type)
            
        if mu_type=='H':
            rho =self.get_rho(t,rho_in,reshape=True,original_L_basis=True)

            if ket_flag:
                exp_val = np.einsum('ij,jil',mu,rho)
            else:
                exp_val = np.einsum('ijl,ji',rho,mu)
            
        else:
            rho =self.get_rho(t,rho_in,reshape=False,original_L_basis=False)
            m_nonzero = rho_in.bool_mask

            # e_mask = self.electric_field_mask(pulse_number,mu_key,conjugate_flag=conjugate_flag)
            # boolean_matrix = boolean_matrix * e_mask
            # overlap_matrix = overlap_matrix * e_mask

            mu,n_nonzero=self.mask_dipole_matrix(boolean_mu,mu,m_nonzero)

            rho = mu.dot(rho)

            rho = self.rho_L_vector_to_matrix(rho,new_manifold_key,
                                              n_nonzero)
            exp_val = np.einsum('iik',rho)
            
        return exp_val
    
    def fl_exp_val(self,t,rho_in):
        rho =self.get_rho(t,rho_in,reshape=True,original_L_basis=True)
        pops = rho[...,0].diagonal()
        if 'all_manifolds' in self.manifolds:
            fl = self.fluorescence_yield['all_manifolds']
        else:
            fl = self.fluorescence_yield[rho_in.manifold_key]
        exp_val = np.dot(fl, pops)
        
        return exp_val
            
    
    def next_order(self,rho_in,*,ket_flag=True,up_flag=True,
                   next_manifold_mask = None,pulse_number = 0):
        """This function connects rho_p to rho_pj^(*) using a DFT convolution algorithm.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold
        
        Return:
            rho_dict (perturbative_container): next-order density matrix
"""
        ### Generic
        tnext_order0 = time.time()
        if ket_flag == up_flag:
            # Rotating term excites the ket and de-excites the bra
            conjugate_flag = False
        else:
            # Counter-rotating term
            conjugate_flag = True

        if rho_in.pulse_number == None:
            old_pulse_time = -np.inf
        else:
            old_pulse_time = self.pulse_times[rho_in.pulse_number]
        pulse_time = self.pulse_times[pulse_number]

        if np.isclose(old_pulse_time,pulse_time):
            simultaneous = rho_in.simultaneous + 1
        else:
            simultaneous = 1

        t = self.efield_times[pulse_number] + pulse_time
        dt = self.dts[pulse_number]
        old_manifold_key = rho_in.manifold_key
        new_manifold_key = self.get_new_manifold_key(old_manifold_key,
                                                     ket_flag,up_flag)

        input_pdc = rho_in.pdc
        output_pdc = input_pdc.copy()
        if conjugate_flag:
            output_pdc[pulse_number][1] += 1
        else:
            output_pdc[pulse_number][0] += 1

        center = self.centers[pulse_number]
        efield = self.efields[pulse_number] * np.exp(-1j*center*t)

        if conjugate_flag:
            efield = np.conjugate(efield)

        if ket_flag:
            efield = 1j * efield
        else:
            efield = -1j * efield

        ### UF2 specific

        mu_type = 'H' if self.conserve_memory else 'L'

        rho,n_nonzero=self.mu_dot_rho(t,rho_in,pulse_number,mu_type=mu_type,
                                         ket_flag=ket_flag,up_flag=up_flag,
                                     next_manifold_mask=next_manifold_mask)
        
        ### Generic
        rho = rho * efield[np.newaxis,:]

        ### UF2 specific
        try:
            ev2 = self.eigenvalues['all_manifolds']
        except KeyError:
            ev2 = self.eigenvalues[new_manifold_key]

        if self.interaction_picture_shift:
            t_exp_2 = t - pulse_time
        else:
            t_exp_2 = t

        exp_factor2_arg = -ev2[n_nonzero,np.newaxis] * t_exp_2[np.newaxis,:]
        exp_factor2 = np.exp(exp_factor2_arg)
        
        if self.interaction_picture_calculations:
            rho = rho * exp_factor2
        else:
            t_exp_2b = t_exp_2 - t_exp_2[0]
            exp_factor2b_arg = ev2[n_nonzero,np.newaxis] * t_exp_2b[np.newaxis,:]
            exp_factor2b = np.exp(exp_factor2b_arg)
            print('Possible problem with non-interaction picture calculations')

        t0 = time.time()

        M = self.efield_times[pulse_number].size

        fft_convolve_fun = self.heaviside_convolve_list[pulse_number].fft_convolve2

        if M == 1:
            pass
        else:
            if self.interaction_picture_calculations:
                rho = fft_convolve_fun(rho,d=dt)
            else:
                rho = fft_convolve_fun(exp_factor2b,rho,d=dt)

        if not self.interaction_picture_calculations:
            rho = rho * exp_factor2

        t1 = time.time()
        self.convolution_time += t1-t0

        rho_out = perturbative_container(t,rho,n_nonzero,pulse_number,
                                new_manifold_key,output_pdc,pulse_time,
                                simultaneous=simultaneous)

        self.next_order_counter += 1
        self.next_order_time += time.time() - tnext_order0

        return rho_out

class Ununsed:

    def mask_negligible_states(self,t,evs):
        """Find eigenstates that have decayed to negligible values and 
            return mask to remove them
"""
        ev_t = t[0] * evs
        inds = np.where(np.real(ev_t) < -7)[0]
        mask = np.ones(evs.size,dtype='bool')
        mask[inds] = False
        return mask



#Standard python libraries
import os
import warnings
import copy
import itertools
import time

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse import csr_matrix, identity, kron
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq
from scipy.integrate import RK45

#Other parts of this code
from ufss.HLG.base_class import DataOrganizer
from ufss.HLG.eigenstates import PolymerVibrations

class RKE_Wavepackets(PolymerVibrations,DataOrganizer):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the hamiltonians of each optical
        manifold

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        num_conv_points (int): number of desired points for linear 
            convolution. Also number of points used to resolve all optical
            pulse shapes
        dt (float): time spacing used to resolve the shape of all optical
            pulses
        initial_state (int): index of initial state for psi^0 NOT IMPLEMENTED YET
        total_num_time_poitns (int): total number of time points used for
            the spectroscopic calculations.

"""
    def __init__(self,file_path,*, num_conv_points=138, dt=0.1,center = 0,
                 initial_state=0, total_num_time_points = 2000):
        self.size = num_conv_points
        # Initialize time array to be used for all desired delay times
        self.t = np.arange(-(total_num_time_points//2),total_num_time_points//2+total_num_time_points%2)*dt
        self.t += self.t[-(self.size//2+1)]
        self.dt = dt
        parameter_file = os.path.join(file_path,'params.yaml')
        super().__init__(parameter_file,mask_by_occupation_num=True)
        self.base_path = file_path

        self.load_params()

        self.set_diagrams_and_manifolds()

        self.set_molecular_dipoles()

        self.set_bottom_eigensystem()

        self.set_H()

        self.zero_hamiltonians()

        self.rtol = 1E-6
        self.atol = 1E-6

        self.time_to_extend = 0
        self.time_for_next_order = 0

        ############### Optical part

        self.set_homogeneous_linewidth(0.05)

        self.efield_t = np.arange(-(num_conv_points//2),num_conv_points//2+num_conv_points%2) * dt
        self.efield_w = 2*np.pi*fftshift(fftfreq(self.efield_t.size,d=dt))

        # Code will not actually function until the following three empty lists are set by the user
        self.efields = [] #initialize empty list of electric field shapes
        self.polarization_sequence = [] #initialize empty polarization sequence
        self.pulse_times = [] #initialize empty list of pulse arrival times
        
        f = fftshift(fftfreq(self.t.size-self.t.size%2,d=self.dt))
        self.w = 2*np.pi*f

        # Define the unitary operator for each manifold in the RWA given the rotating frequency center
        self.recenter(new_center = center)

    def set_diagrams_and_manifolds(self):
        if len(self.energies) == 2:
            self.diagrams = ['GSB', 'SE']
            self.manifolds = ['GSM','SEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE']}
        elif len(self.energies) == 3:
            self.diagrams = ['GSB','SE','ESA']
            self.manifolds = ['GSM','SEM','DEM']
            self.diagrams_by_manifold = {'GSM':['GSB'],'SEM':['SE','ESA']}

    def set_H(self,*,truncation_size = None):
        if truncation_size:
            self.truncation_size = truncation_size
            self.set_vibrations()
        self.H0 = self.manifold_hamiltonian(0)
        self.H1 = self.manifold_hamiltonian(1)
        
        if 'DEM' in self.manifolds:
            self.H2 = self.manifold_hamiltonian(2)

    def make_H_dense(self):
        self.H0 = self.H0.toarray()
        self.H1 = self.H1.toarray()
        if 'DEM' in self.manifolds:
            self.H2 = self.H2.toarray()

    def dH0(self,t,psi_in):
        return -1j*self.H0.dot(psi_in)

    def dH1(self,t,psi_in):
        return -1j*self.H1.dot(psi_in)

    def dH2(self,t,psi_in):
        return -1j*self.H2.dot(psi_in)

    def RKU_onetimestep(self,psi0,t0,tf,*,manifold_num = 0):
        dH = [self.dH0,self.dH1,self.dH2]
        fun = dH[manifold_num]
        rk45 = RK45(fun,t0,psi0,tf,atol=self.atol,rtol=self.rtol,vectorized=True)
        while rk45.t < tf:
            rk45.step()
        psi_final = rk45.y
        return psi_final

    def RKU(self,psi0,t0,tvalues,*,manifold_num = 0):
        t_last = t0
        states = np.zeros((psi0.size,tvalues.size),dtype='complex')
        psi_last = psi0
        for i in range(tvalues.size):
            t_next = tvalues[i]
            psi_next = self.RKU_onetimestep(psi_last,t_last,t_next,
                                               manifold_num=manifold_num)
            states[:,i] = psi_next
            t_last = t_next
            psi_last = psi_next

        return states

    def set_molecular_dipoles(self,*,dipoles = None):
        """Load molecular dipoles from params file, or override with input
dipoles - must be a numpy ndarray, with shape (n,3) where n is the number of sites"""
        if type(dipoles) is np.ndarray:
            self.molecular_dipoles = dipoles
        else:
            self.molecular_dipoles = np.array(self.params['dipoles'],dtype='float')

        self.set_single_to_double_dipole_matrix()

    def set_single_to_double_dipole_matrix(self):
        """Given a set of dipoles for transitions from the ground to the
singly excited states, constructs the dipole transitions that take the
system from the singly excited states to the various doubly excited states
"""
        singly_excited = np.arange(self.molecular_dipoles.shape[0])
        doubly_excited = list(itertools.combinations(singly_excited,2))
        mat = np.zeros((len(singly_excited),len(doubly_excited),3))
        for i in range(len(singly_excited)):
            for j in range(len(doubly_excited)):
                tup = doubly_excited[j]
                if i == tup[0]:
                    mat[i,j,:] = self.molecular_dipoles[singly_excited[tup[1]]]
                elif i == tup[1]:
                    mat[i,j,:] = self.molecular_dipoles[singly_excited[tup[0]]]
        self.molecular_dipoles_SEM_to_DEM = mat

    def set_bottom_eigensystem(self):
        """Calculate or load lowest ground-state eigenvalue and eigenvector,
            and lowest singly excited eigenvalue.
            Note: would like to expand this to allow initial conditions other 
            than bottom ground state
"""
        try:
            self.load_bottom_eigensystem()
        except:
            self.calculate_bottom_eigensystem()
            self.save_bottom_eigensystem()

        self.setup_eigensystem()

    def calculate_bottom_eigensystem(self):
        ge0, gv0 = self.eigs(1,0)
        se0, sv0 = self.eigs(1,1)
        self.eigenvalues = [ge0,se0]
        self.eigenvectors = gv0

    def setup_eigensystem(self):
        psi0 = self.eigenvectors[:,0]
        psi0_extension = np.ones((psi0.size,self.t.size),dtype='complex')
        psi0 = psi0[:,np.newaxis] * psi0_extension
        self.psi0 = {'psi':psi0,'manifold_num':0,'bool_mask':np.array([None])}
        ge0 = self.eigenvalues[0][0]
        se0 = self.eigenvalues[1][0]
        self.ground_ZPE = ge0
        self.ground_to_excited_transition = se0 - ge0

    def zero_hamiltonians(self):
        # Subtract out GSM zpe
        self.H0 -= identity(self.H0.shape[0]) * self.ground_ZPE
        self.H1 -= identity(self.H1.shape[0]) * self.ground_ZPE
        if 'DEM' in self.manifolds:
            self.H2 -= identity(self.H2.shape[0]) * self.ground_ZPE

        # Default RWA is the gap from lowest energy ground state
        # and lowest energy singly excited state
        self.H1 -= identity(self.H1.shape[0]) * self.ground_to_excited_transition
        if 'DEM' in self.manifolds:
            self.H2 -= identity(self.H2.shape[0]) * 2 * self.ground_to_excited_transition

    def save_bottom_eigensystem(self):
        np.savez(os.path.join(self.base_path,'bottom_eigenvalues'),GSM=self.eigenvalues[0],SEM=self.eigenvalues[1])
        np.save(os.path.join(self.base_path,'bottom_eigenvectors'),self.eigenvectors)

    def load_bottom_eigensystem(self):
        arch = np.load(os.path.join(self.base_path,'bottom_eigenvalues.npz'))
        self.eigenvalues = [arch['GSM'],arch['SEM']]
        self.eigenvectors = np.load(os.path.join(self.base_path,'bottom_eigenvectors.npy'))

    def set_homogeneous_linewidth(self,gamma):
        self.gamma = gamma

    def recenter(self,new_center = 0):
        self.H1 -= identity(self.H1.shape[0]) * new_center
        if 'DEM' in self.manifolds:
            self.H2 -= identity(self.H2.shape[0]) * new_center

    def get_closest_index_and_value(self,value,array):
        """Given an array and a desired value, finds the closest actual value
            stored in that array, and returns that value, along with its 
            corresponding  array index
"""
        index = np.argmin(np.abs(array - value))
        value = array[index]
        return index, value

    def extend_wavefunction(self,psi_dict,pulse_start_ind,pulse_end_ind,
                            gamma_end_ind = None):
        """Perturbative wavefunctions are calculated only during the time where the given pulse
            is non-zero.  This function extends the wavefunction beyond those bounds by taking 
            all values before the interaction to be zero, and using RK45 method to extend the 
            wavefunction forward
"""
        t_start = time.time()

        pulse_t_slice = slice(pulse_start_ind, pulse_end_ind,1)

        psi = psi_dict['psi']

        m_num = psi_dict['manifold_num']

        total_psi = np.zeros((psi.shape[0],self.t.size),dtype='complex')
        total_psi[:,pulse_t_slice] = psi
        psi_last = total_psi[:,pulse_end_ind-1]
        t_last = self.t[pulse_end_ind-1]
        total_psi[:,pulse_end_ind:gamma_end_ind] = self.RKU(psi_last,t_last,
                                                            self.t[pulse_end_ind:gamma_end_ind],
                                                            manifold_num = m_num)

        psi_dict['psi'] = total_psi
        self.time_to_extend += time.time() - t_start
        return psi_dict

    def set_polarization_sequence(self,polarization_list,*,reset_psis=True):
        """Sets the sequences used for either parallel or crossed pump and probe
        
        Args:
            polarization_list (list): list of four strings, can be 'x' or 'y'
        Returns:
            None: sets the attribute polarization sequence
"""

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        pol_options = {'x':x,'y':y,'z':z}

        self.polarization_sequence = [pol_options[pol] for pol in polarization_list]

        if reset_psis:
            self.psis = dict()


    def dipole_matrix(self,starting_manifold_num,next_manifold_num,pulse_number):
        """Calculates the dipole matrix that connects from one 
            manifold to the next, using the known dipole moments and the efield 
            polarization, determined by the pulse number.
"""
        pol = self.polarization_sequence[pulse_number]
        upper_manifold_num = max(starting_manifold_num,next_manifold_num)
        
        if abs(starting_manifold_num - next_manifold_num) != 1:
            warnings.warn('Can only move from manifolds 0 to 1 or 1 to 2')
            return None
        
        # Condon approximation
        vib_identity = identity(self.H0.shape[0])
        
        if upper_manifold_num == 1:
            d_vec = self.molecular_dipoles.dot(pol)
            d_mat = d_vec[:,np.newaxis]
            overlap_matrix = kron(d_mat,vib_identity)
            
        elif upper_manifold_num == 2:
            d_mat = self.molecular_dipoles_SEM_to_DEM.dot(pol)
            overlap_matrix = kron(d_mat.T,vib_identity)

        if starting_manifold_num > next_manifold_num:
            # Take transpose if transition is down rather than up
            overlap_matrix = np.conjugate(overlap_matrix.T)

        return overlap_matrix.tocsr()

    def next_order(self,psi_in_dict,manifold_change,*,pulse_number = 0,gamma = 0,
                   new_manifold_mask = None):
        """This function connects psi_p to psi+pj^(*)

        Args:
            psi_in_dict (dict): input wavefunction dictionary
            manifold_change (int): is either +/-1 (up or down)
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
        
        Return:
            psi_dict (dict): next-order wavefunction
"""
        t_start = time.time()
        pulse_time = self.pulse_times[pulse_number]
        pulse_time_ind = np.argmin(np.abs(self.t - pulse_time))

        pulse_start_ind = pulse_time_ind - self.size//2
        pulse_end_ind = pulse_time_ind + self.size//2 + self.size%2

        t_slice = slice(pulse_start_ind, pulse_end_ind,1)
        t = self.t[t_slice]

        if gamma != 0:
            gamma_end_ind = pulse_time_ind + int(6.91/self.gamma/self.dt)
        else:
            gamma_end_ind = None
        
        psi_in = psi_in_dict['psi'][:,t_slice].copy()
        
        starting_manifold_num = psi_in_dict['manifold_num']
        next_manifold_num = starting_manifold_num + manifold_change
        
        dipole_matrix = self.dipole_matrix(starting_manifold_num,next_manifold_num,
                                            pulse_number)
        
        if pulse_number == 'impulsive':
            # This is broken!
            psi_imp = dipole_matrix.dot(psi_in[:,self.size//2])
            psi = np.zeros((psi_imp.size,t.size),dtype='complex')
            psi[:,self.size//2] = psi_imp
            t_last = t[self.size//2]
            t_request = t[self.size//2+1:]
            psi[:,self.size//2+1:] = self.RKU(psi_imp,t_last,t_request,
                                              manifold_num=next_manifold_num)
        else:
            if next_manifold_num > starting_manifold_num:
                efield = self.efields[pulse_number]
            else:
                efield = np.conjugate(self.efields[pulse_number])
            psi = np.zeros((dipole_matrix.shape[0],t.size),dtype='complex')
            psi[:,0] = 1j*dipole_matrix.dot(psi_in[:,0]) * efield[0] * self.dt
            for i in range(1,efield.size):
                psi[:,i] = self.RKU_onetimestep(psi[:,i-1],t[i-1],t[i],
                                                manifold_num = next_manifold_num)
                psi[:,i] += 1j*dipole_matrix.dot(psi_in[:,i]) * efield[i] * self.dt
            

        psi_dict = {'psi':psi,'manifold_num':next_manifold_num,'bool_mask':np.array([None])}

        self.time_for_next_order += time.time() - t_start

        psi_dict = self.extend_wavefunction(psi_dict,pulse_start_ind,pulse_end_ind,
                                            gamma_end_ind = gamma_end_ind)

        return psi_dict

    def up(self,psi_in_dict,*,pulse_number = 0,gamma=0,new_manifold_mask = None):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            psi_in_dict (dict): input wavefunction dictionary
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            gamma (float): optical dephasing (only use with final interaction)
            new_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (dict): output from method next_order
"""

        return self.next_order(psi_in_dict,1,gamma=gamma,
                               pulse_number = pulse_number)

    def down(self,psi_in_dict,*,pulse_number = 0,gamma=0,new_manifold_mask = None):
        """This method connects psi_p to psi_pj^* where the next order psi
            is one manifold below the current manifold.

        Args:
            psi_in_dict (dict): input wavefunction dictionary
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            gamma (float): optical dephasing (only use with final interaction)
            new_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (dict): output from method next_order
"""

        return self.next_order(psi_in_dict,-1,gamma=gamma,
                               pulse_number = pulse_number)

    ### Tools for taking the expectation value of the dipole operator with perturbed wavepackets
    
    def dipole_down(self,psi_in_dict,*,pulse_number = -1):
        """This method is similar to the method down, but does not
involve the electric field shape or convolutions. It is the action of 
the dipole operator on a ket without TDPT effects.  It also includes
the dot product of the final electric field polarization vector."""
        psi_in = psi_in_dict['psi']
        starting_manifold_num = psi_in_dict['manifold_num']
        next_manifold_num = starting_manifold_num - 1
        
        # This function is always used as the final interaction to
        # produce the polarization field, which is a vector quantity
        # However we assume that we will calculate a signal, which
        # invovles the dot product of the polarization field with the
        # local oscillator vector. We do this now to avoid carrying
        # around the cartesian coordinates of the polarization field
        
        overlap_matrix = self.dipole_matrix(starting_manifold_num,next_manifold_num,
                                                            pulse_number = pulse_number)

        psi = overlap_matrix.dot(psi_in)

        psi_dict = {'psi':psi,'manifold_num':next_manifold_num,'bool_mask':np.array([None])}

        return psi_dict

    def dipole_expectation(self,bra_dict_original,ket_dict_original,*,pulse_number = -1):
        """Computes the expectation value of the two wavefunctions with respect 
            to the dipole operator.  Both wavefunctions are taken to be kets, and the one named 'bra' is
converted to a bra by taking the complex conjugate."""
        pulse_time = self.pulse_times[pulse_number]
        pulse_time_ind = np.argmin(np.abs(self.t - pulse_time))

        pulse_start_ind = pulse_time_ind - self.size//2
        gamma_end_ind = pulse_time_ind + int(6.91/self.gamma/self.dt)

        # The signal is zero before the final pulse arrives, and persists
        # until it decays. Therefore we avoid taking the sum at times
        # where the signal is zero. This is captured by t_slice
        t_slice = slice(pulse_start_ind, gamma_end_ind,None)

        bra_in = bra_dict_original['psi'][:,t_slice].copy()
        ket_in = ket_dict_original['psi'][:,t_slice].copy()
        
        manifold1_num = bra_dict_original['manifold_num']
        manifold2_num = ket_dict_original['manifold_num']

        bra_dict = {'manifold_num':manifold1_num,'psi':bra_in}
        ket_dict = {'manifold_num':manifold2_num,'psi':ket_in}

        if np.abs(manifold1_num - manifold2_num) != 1:
            warnings.warn('Dipole only connects manifolds 0 to 1 or 1 to 2')
            return None

        if manifold1_num > manifold2_num:
            bra_dict = self.dipole_down(bra_dict,
                                        pulse_number = pulse_number)
        else:
            ket_dict = self.dipole_down(ket_dict,
                                        pulse_number = pulse_number)

        bra = bra_dict['psi']
        ket = ket_dict['psi']

        exp_val = np.sum(np.conjugate(bra) * ket,axis=0)
        
        # Initialize return array with zeros
        ret_val = np.zeros(self.t.size,dtype='complex')
        # set non-zero values using t_slice
        ret_val[t_slice] = exp_val
        return ret_val

    def integrated_dipole_expectation(self,bra_dict_original,ket_dict_original,*,pulse_number = -1):
        """Given two wavefunctions, this computes the expectation value of the two with respect 
to the dipole operator.  Both wavefunctions are taken to be kets, and the one named 'bra' is
converted to a bra by taking the complex conjugate.  This assumes that the signal will be
frequency integrated."""
        pulse_time = self.pulse_times[pulse_number]
        pulse_time_ind = np.argmin(np.abs(self.t - pulse_time))

        pulse_start_ind = pulse_time_ind - self.size//2
        pulse_end_ind = pulse_time_ind + self.size//2 + self.size%2
        
        # The signal is zero before the final pulse arrives, and persists
        # until it decays. However, if no frequency information is
        # required, fewer time points are needed for this t_slice
        t_slice = slice(pulse_start_ind, pulse_end_ind,None)

        bra_in = bra_dict_original['psi'][:,t_slice].copy()
        ket_in = ket_dict_original['psi'][:,t_slice].copy()
        
        
        manifold1_num = bra_dict_original['manifold_num']
        manifold2_num = ket_dict_original['manifold_num']

        bra_dict = {'manifold_num':manifold1_num,'psi':bra_in,'bool_mask':np.array([None])}
        ket_dict = {'manifold_num':manifold2_num,'psi':ket_in,'bool_mask':np.array([None])}

        if np.abs(manifold1_num - manifold2_num) != 1:
            warnings.warn('Dipole only connects manifolds 0 to 1 or 1 to 2')
            return None

        if manifold1_num > manifold2_num:
            bra_new_mask = ket_dict['bool_mask']
            bra_dict = self.dipole_down(bra_dict,
                                        pulse_number = pulse_number)
        else:
            ket_new_mask = bra_dict['bool_mask']
            ket_dict = self.dipole_down(ket_dict,
                                        pulse_number = pulse_number)

        bra = bra_dict['psi']
        ket = ket_dict['psi']

        exp_val = np.sum(np.conjugate(bra) * ket,axis=0)
        return exp_val

    def polarization_to_signal(self,P_of_t_in,*,return_polarization=False,
                               local_oscillator_number = -1):
        """This function generates a frequency-resolved signal from a polarization field
local_oscillator_number - usually the local oscillator will be the last pulse in the list self.efields"""
        pulse_time = self.pulse_times[local_oscillator_number]
        if self.gamma != 0:
            exp_factor = np.exp(-self.gamma * (self.t-pulse_time))
            P_of_t_in *= exp_factor
        P_of_t = P_of_t_in
        if return_polarization:
            return P_of_t

        if local_oscillator_number == 'impulsive':
            efield = np.exp(1j*self.w*(pulse_time))
        else:
            pulse_time_ind = np.argmin(np.abs(self.t - pulse_time))

            pulse_start_ind = pulse_time_ind - self.size//2
            pulse_end_ind = pulse_time_ind + self.size//2 + self.size%2

            t_slice = slice(pulse_start_ind, pulse_end_ind,None)
            efield = np.zeros(self.t.size,dtype='complex')
            efield[t_slice] = self.efields[local_oscillator_number]
            efield = fftshift(ifft(ifftshift(efield)))*len(P_of_t)*(self.t[1]-self.t[0])/np.sqrt(2*np.pi)

        if P_of_t.size%2:
            P_of_t = P_of_t[:-1]
            efield = efield[:len(P_of_t)]

        P_of_w = fftshift(ifft(ifftshift(P_of_t)))*len(P_of_t)*(self.t[1]-self.t[0])/np.sqrt(2*np.pi)
        
        signal = np.imag(P_of_w * np.conjugate(efield))
        return signal

    def polarization_to_integrated_signal(self,P_of_t,*,
                                          local_oscillator_number = -1):
        """This function generates a frequency-integrated signal from a polarization field
local_oscillator_number - usually the local oscillator will be the last pulse in the list self.efields
"""
        pulse_time = self.pulse_times[local_oscillator_number]
        pulse_time_ind = np.argmin(np.abs(self.t - self.delay_time))

        pulse_start_ind = pulse_time_ind - self.size//2
        pulse_end_ind = pulse_time_ind + self.size//2 + self.size%2

        t_slice = slice(pulse_start_ind, pulse_end_ind,1)
        t = self.t[t_slice]
        if self.gamma != 0:
            exp_factor = np.exp(-self.gamma * (t-pulse_time))
            P_of_t *= exp_factor

        if local_oscillator_number == 'impulsive':
            signal = P_of_t[self.size//2]
        else:
            efield = self.efields[local_oscillator_number]
            signal = np.trapz(np.conjugate(efield)*P_of_t,x=t)
        
        return np.imag(signal)

    



import unittest
import numpy as np
import os
from scipy.io import loadmat

import ufss

def calculate_spectrum_with_UF2():
    folder = os.path.join('fixtures','so_comparison_dimer')
    ufss.HLG.run(folder)

    ta = ufss.Wavepackets(os.path.join(folder,'closed'),
                              detection_type='integrated_polarization')

    # defining the optical pulses in the RWA
    M = 41 # number of points used to resolve optical pulses
    Delta = 10 # pulse interval
    t = np.linspace(-Delta/2,Delta/2,num=M)
    dt = t[1] - t[0]

    sigma = 1
    ef = ufss.gaussian(t,sigma)

    ta.set_polarization_sequence(['x','x'])

    ta.set_efields([t,t],[ef,ef],[0,0],[(1,1),(1,0)])

    # ta.set_t(gamma_dephasing)

    T = np.arange(0.6,112,0.6)

    ta.set_pulse_delays([T])

    sig = ta.calculate_signal_all_delays()

    return sig

def L2_norm(a,b):
    """Computes L2 norm of two nd-arrays, taking the b as the reference"""
    return np.sqrt(np.sum(np.abs(a-b)**2)/np.sum(np.abs(b)**2))
    

class test_against_matlab(unittest.TestCase):
    """This test verifies that UFSS produces the same frequency-integrated
transient absorption spectrum as is generated using the code released with
the Ultrafast Spectroscopy book 
(https://iopscience.iop.org/book/978-0-750-31062-8, see supplementary 
material). The spectrum contained in the file 
fixtures/so_comparison_dimer/MATLAB_SOP_comparison.mat was obtained by 
running the downloaded code and using the parameters contained in the file
fixtures/so_comparison_dimer/MATLAB_SOP_comparison_params.mat.

This test serves to simultaneously test the diagram generator, the 
Hamiltonian generator, and UF2, since the comparison method uses
a different method for representing the Hamiltonian, and a different method
for propagating the wavefunctions and including the pulse shapes. The other
code Feynman diagrams that are written by hand. This does NOT test the 
Liouvillian generator or the open UF2 algorithm, since the MATLAB code only
works for closed systems.
"""
    def test(self):
        uf2_sig = calculate_spectrum_with_UF2()
        matlab_sig_path = os.path.join('fixtures','so_comparison_dimer',
                                       'MATLAB_SOP_comparison.mat')
        matlab_sig = loadmat(matlab_sig_path)['signal_T'][0,:]
        diff = L2_norm(-2*uf2_sig,matlab_sig)
        self.assertTrue(diff < 0.01)

if __name__ == '__main__':
    unittest.main()



#Standard python libraries
import copy
import os
import time

#Dependencies
import numpy as np
import warnings
import matplotlib.pyplot as plt

#UF2
from .plotting_tools import Plotter
from .signal_processing import SignalProcessing

class AnalyzeTransientAbsorption(SignalProcessing,Plotter):

    def __init__(self,parameter_file_path,*,load_file_name='default'):
        self.base_path = parameter_file_path
        self.load(load_file_name=load_file_name)

    def load(self,*,load_file_name='default'):
        if load_file_name == 'default':
            load_name = os.path.join(self.base_path,'TA_spectra_iso_ave.npz')
        else:
            load_name = os.path.join(self.base_path,load_file_name)
        arch = np.load(load_name)
        self.signal = arch['signal']
        if 'delay_times' in arch.keys():
            self.delay_times = arch['delay_times']
        elif 't0' in arch.keys():
            self.delay_times = arch['t0']
        elif 'T' in arch.keys():
            self.delay_times = arch['T']
        else:
            raise Exception('Could not load delay times from signal archive')
        try:
            self.w = arch['frequencies']
        except KeyError:
            self.w = arch['wt']
        try:
            self.center = arch['pulse_center']
        except KeyError:
            warnings.warn('Pulse center was not saved in archive, setting center = 0')
            self.center = 0

    def get_closest_index_and_value(self,value,array):
        """Given an array and a desired value, finds the closest actual value
            stored in that array, and returns that value, along with its 
            corresponding array index
"""
        index = np.argmin(np.abs(array - value))
        value = array[index]
        return index, value

    def phase_diff(self,signal,w1,w2,wT):
        w1ind, w1 = self.get_closest_index_and_value(w1,self.w)
        w2ind, w2 = self.get_closest_index_and_value(w2,self.w)
        wTind, wT = self.get_closest_index_and_value(wT,self.wT)
        signal_ft = self.ft_axis1(signal)
        return self.phase_diff_by_inds(signal_ft,w1ind,w2ind,wTind)

    def mag_ratio(self,signal,w1,w2,wT):
        w1ind, w1 = self.get_closest_index_and_value(w1,self.w)
        w2ind, w2 = self.get_closest_index_and_value(w2,self.w)
        wTind, wT = self.get_closest_index_and_value(wT,self.wT)
        signal_ft = self.ft_axis1(signal)
        return self.mag_ratio_by_inds(signal_ft,w1ind,w2ind,wTind)

    def integrated_signal(self,*,frequency_range='all'):
        return self.integrate_TA(self.w, self.signal,
                                 x_range = frequency_range)

    def integrated_ft(self,*,frequency_range='all',delay_time_start = -1,delay_time_stop = 300,gamma_T = 0):
        delay_time_indices = np.where((self.delay_times > delay_time_start)
                                      & (self.delay_times < delay_time_stop))[0]
        delay_times = self.delay_times[delay_time_indices]
        signal = self.signal[delay_time_indices,:].copy()
        if gamma_T != 0:
            signal *= np.exp(-gamma_T*np.abs(delay_times))[:,np.newaxis]
        w_T, signal = self.ft1D(delay_times,signal,axis=0)
        return w_T,self.integrate_TA(self.w, signal,
                                 x_range = frequency_range)

    def find_node(self):
        """Finds the location of the vibrational node
"""
        zeros = np.zeros(self.signal.shape[0])
        T, sig = self.remove_DC(self.delay_times,self.signal,axis=0)
        for i in range(zeros.size):
            zero = self.find_zero_closest_to(self.w,np.real(sig[i,:]))
            zeros[i] = zero
        return zeros

    def find_average_node(self,*,max_delay_time = 50):
        nodes = self.find_node()
        Tind, Tmax = self.get_closest_index_and_value(max_delay_time,self.delay_times)
        nodes_to_fit = nodes[:Tind]
        finite_indices = np.where(np.isfinite(nodes_to_fit))[0]
        return np.mean(nodes_to_fit[finite_indices])

    def plot_node(self):
        zeros = self.find_node()
        plt.plot(self.delay_times,zeros,'--r')
        
    def find_node_wT(self,wT):
        w_T, sig = self.ft1D(self.delay_times,self.signal,axis=0)
        wTind, wT = self.get_closest_index_and_value(wT,w_T)
        imag_zero = self.find_zero_closest_to(self.w,np.imag(sig[wTind,:]))
        real_zero = self.find_zero_closest_to(self.w,np.real(sig[wTind,:]))
        return real_zero,imag_zero

    def kappa_measure(self,wT,delay_time_start = 1,gamma_T = 0,node_position=0):
        ave_node = self.find_average_node()
        m = self.kappa_measure_manual_node(wT,delay_time_start = delay_time_start,
                                           gamma_T = gamma_T,node_position=ave_node)
        return m

    def kappa_measure_manual_node(self,wT,delay_time_start = 1,gamma_T = 0,node_position=0):
        w_T, neg_integral = self.integrated_ft(frequency_range=[-np.inf,node_position+0.0001],delay_time_start=delay_time_start,
                                               gamma_T=gamma_T)
        w_T, pos_integral = self.integrated_ft(frequency_range=[node_position-0.0001,np.inf],delay_time_start=delay_time_start,
                                               gamma_T = gamma_T)
        wTind, wT = self.get_closest_index_and_value(wT,w_T)
        neg = np.abs(neg_integral[wTind])
        pos = np.abs(pos_integral[wTind])
        ave = (neg + pos)/2
        diff = (pos - neg)/2
        return diff/ave

    def kappa_measure_old(self,wT,delay_time_start = 1,gamma_T = 0):
        w_T, neg_integral = self.integrated_ft(frequency_range=[-np.inf,0.0001],delay_time_start=delay_time_start,
                                               gamma_T=gamma_T)
        w_T, pos_integral = self.integrated_ft(frequency_range=[-0.0001,np.inf],delay_time_start=delay_time_start,
                                               gamma_T = gamma_T)
        wTind, wT = self.get_closest_index_and_value(wT,w_T)
        neg = neg_integral[wTind]
        pos = pos_integral[wTind]
        if np.abs(pos) > np.abs(neg):
            k = +1
        else:
            k = -1
        ave = np.abs(neg + pos)/2
        diff = np.abs(neg - pos)/2
        return k*diff/ave

    def add_gaussian_linewidth(self,sigma):
        self.old_signal = self.signal.copy()

        t,sig_t = self.ft1D(self.w,self.old_signal,axis=-1)
        sig_t = sig_t * np.exp(-t**2/(2*sigma**2))[np.newaxis,:]
        w,sig_w = self.ift1D(t,sig_t,axis=-1)
        self.signal = sig_w

    def plot(self,*,remove_DC = True,frequency_range=[-3,3]):
        self.plotTA(subtract_DC = remove_DC, create_figure=True,
                    color_range = 'auto',draw_colorbar = True,save_fig=False,
                    frequency_range=frequency_range)
        # if remove_DC:
        #     T, sig = self.remove_DC(self.delay_times,self.signal,axis=0)
        # else:
        #     sig = self.signal
        # ufss.signals.plot2D(self.delay_times,sig,part='real')



#Standard python libraries
import os

#Dependencies
import numpy as np
import yaml
import matplotlib.pyplot as plt
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq
from scipy.optimize import brentq as sbrentq
from scipy.interpolate import interp1d as sinterp1d


class SignalProcessing(object):
    @staticmethod
    def subtract_DC(signal,return_ft = False, axis = 1):
        """Use discrete fourier transform to remove the DC component of a 
            signal.
        Args:
            signal (np.ndarray): real signal to be processed
            return_ft (bool): if True, return the Fourier transform of the 
                              input signal
            axis (int): axis along which the fourier trnasform is to be taken
"""
        sig_fft = fft(ifftshift(signal,axes=(axis)),axis=axis)
        
        nd_slice = [slice(None) for i in range(len(sig_fft.shape))]
        nd_slice[axis] = slice(0,1,1)
        nd_slice = tuple(nd_slice)
        sig_fft[nd_slice] = 0
        
        if not return_ft:
            sig = fftshift(ifft(sig_fft),axes=(axis))
        else:
            sig = sig_fft
        return sig

    @staticmethod
    def remove_DC(x,y,*,axis=0):
        """Redundant with subtract_DC
"""
        k,f = SignalProcessing.ft1D(x,y,axis=axis,zero_DC=True)
        x2,y2 = SignalProcessing.ift1D(k,f,axis=axis,zero_DC=False)
        if np.isclose(x[0],0):
            y2 = ifftshift(y2,axes=(axis))
        return x, y2
            

    def integrate_TA(self,x,signal,*,x_range = 'all'):
        if x_range == 'all':
            pass
        else:
            inds = np.where((x > x_range[0]) & (x < x_range[1]))[0]
            x = x[inds]
            signal = signal[:,inds]
        return np.trapz(signal,x=x,axis=1)
    
    @staticmethod
    def ft1D(x,y,*,axis=0,zero_DC=False):
        """Takes in x and y = y(x), and returns k and the Fourier transform 
        of y(x) -> f(k) along a single (1D) axis
        Handles all of the annoyances of fftshift and ifftshift, and gets the 
        normalization right
        Args:
            x (np.ndarray) : independent variable, must be 1D
            y (np.ndarray) : dependent variable, can be nD
        Kwargs:
            axis (int) : which axis to perform FFT
            zero_DC (bool) : if true, sets f(0) = 0
    """
        dx = x[1]-x[0]
        k = fftshift(fftfreq(x.size,d=dx))*2*np.pi
        fft_norm = dx

        shifted_x = ifftshift(x)
        if np.isclose(shifted_x[0],0):
            f = fft(ifftshift(y,axes=(axis)),axis=axis)*fft_norm
        else:
            f = fft(y,axis=axis)*fft_norm

        if zero_DC:
            nd_slice = [slice(None) for i in range(len(f.shape))]
            nd_slice[axis] = slice(0,1,1)
            nd_slice = tuple(nd_slice)
            f[nd_slice] = 0

        f = fftshift(f,axes=(axis))

        return k, f

    @staticmethod
    def ift1D(k,f,*,axis=0,zero_DC=False):
        """Takes in k and f = f(k), and returns x and the discrete Fourier 
        transform of f(k) -> y(x).
        Handles all of the annoyances of fftshift and ifftshift, and gets the 
        normalization right
        Args:
            x (np.ndarray): independent variable
            y (np.ndarray): dependent variable
        Kwargs:
            axis (int) : which axis to perform FFT
    """
        dk = k[1]-k[0]
        x = fftshift(fftfreq(k.size,d=dk))*2*np.pi
        ifft_norm = dk*k.size/(2*np.pi)

        shifted_k = ifftshift(k)
        if np.isclose(shifted_k[0],0):
            y = ifft(ifftshift(f,axes=(axis)),axis=axis)*ifft_norm
        else:
            y = ifft(f,axis=axis)*ifft_norm

        if zero_DC:
            nd_slice = [slice(None) for i in range(len(y.shape))]
            nd_slice[axis] = slice(0,1,1)
            nd_slice = tuple(nd_slice)
            y[nd_slice] = 0

        y = fftshift(y,axes=(axis))

        return x, y

    def phase_2d(self,signal,ind0,ind1):
        val = signal[ind0,ind1]
        phase = np.arctan2(np.real(val),np.imag(val))
        return phase

    def phase_diff_by_inds(self,signal,w1_ind,w2_ind,wT_ind):
        ph1 = self.phase_2d(signal,w1_ind,wT_ind)
        ph2 = self.phase_2d(signal,w2_ind,wT_ind)
        diff = ((ph1-ph2 + np.pi) % (2*np.pi) - np.pi)
        return diff

    def mag_ratio_by_inds(self,signal,w1_ind,w2_ind,wT_ind):
        mag1 = np.abs(signal[w1_ind,wT_ind])
        mag2 = np.abs(signal[w2_ind,wT_ind])
        return mag1/mag2

    def integrated_ft(self,delay_time_start = 1,delay_time_stop = 300):
        delay_time_indices = np.where((self.delay_times > delay_time_start) & (self.delay_times < delay_time_stop))[0]
        delay_times = self.delay_times[delay_time_indices]
        sig = self.signal[delay_time_indices,:]
        integrated = np.trapz(sig,x=self.TA.w,axis=-1)
        w_T = fftshift(fftfreq(delay_times.size,d=(delay_times[1] - delay_times[0])))*2*np.pi
        integrated_fft = fft(integrated)
        integrated_fft[0] = 0
        integrated_fft = fftshift(integrated_fft)
        return w_T, integrated_ft

    def find_all_zeros(self,x,arr):
        inds = np.where(arr[:-1] * arr[1:] < 0)[0]
        zeros = np.zeros(inds.size)
        for i in range(inds.size):
            ind = inds[i]
            zero = (x[ind] + x[ind+1])/2
            zeros[i] = zero
        return zeros

    def find_zero_closest_to(self,x,arr,*,ref=0):
        zeros = self.find_all_zeros(x,arr)
        try:
            min_ind = np.argmin(np.abs(zeros-ref))
            zero = zeros[min_ind]
        except ValueError:
            zero = np.nan
        return zero
        
    def find_zero_brentq(self,x,arr,domain=[-2.2,2.2]):
        """Given an input 1d array, extrapolate a zero crossing
"""
        y = sinterp1d(x,arr)
        try:
            zero = sbrentq(y,*domain)
        except:
            zero = np.nan
        return zero
        



#Standard python libraries
import os
import warnings
import copy
import time
import itertools
import functools

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import numpy.polynomial.chebyshev as npch
import matplotlib.pyplot as plt
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq
from scipy.interpolate import interp1d as sinterp1d
import scipy
from scipy.sparse import csr_matrix, issparse

from ufss.perturbative_calculations import UF2OpenEngine, UF2ClosedEngine
from ufss.perturbative_calculations import RKOpenEngine, RKClosedEngine
from ufss.perturbative_calculations import ChebPoly

class CalculateSignals:

    def __init__(self,*,detection_type = 'polarization'):
        
        self.undersample_factor = 1

        self.gamma_res = 6.91
        
        self.rho_to_polarization_time = 0
        self.add_lorentzian_linewidth = False
        if detection_type == 'polarization':
            self.rho_to_signal = self.polarization_detection_rho_to_signal
            self.return_complex_signal = False
            
        elif detection_type == 'complex_polarization':
            self.rho_to_signal = self.polarization_detection_rho_to_signal
            self.return_complex_signal = True
            
        elif detection_type == 'integrated_polarization':
            self.rho_to_signal = self.integrated_polarization_detection_rho_to_signal
            self.return_complex_signal = False
            
        elif detection_type == 'fluorescence':
            self.rho_to_signal = self.fluorescence_detection_rho_to_signal
            self.return_complex_signal = False

    def set_pulse_delays(self,all_delays):
        """Must be a list of numpy arrays, where each array is a
            list of delay times between pulses
"""
        self.all_pulse_delays = all_delays
        num_delays = len(self.all_pulse_delays)
        num_pulses = len(self.efields)
        
        if num_delays == num_pulses - 1:
            pass
        elif (num_delays == num_pulses - 2
              and 'polarization' in self.detection_type):
            # If there is a local oscillator, it arrives simultaneously with
            # the last pulse by default
            self.all_pulse_delays.append(np.array([0]))
        elif num_delays <= num_pulses -2:
            raise Exception('There are not enough delay times')
        elif num_delays >= num_pulses:
            raise Exception('There are too many delay times')

    def get_initial_and_final_signal_shapes(self,all_delay_combinations):
        # for L interacting pulses, there should be L-1 delays
        final_shape = [delays.size for delays in self.all_pulse_delays]
        if self.detection_type == 'polarization' or self.detection_type == 'complex_polarization':
            initial_shape = (len(all_delay_combinations),self.w.size)
            
            if len(final_shape) == self.pdc.shape[0]:
                # get rid of the "delay" between the last pulse and the local oscillator
                final_shape[-1] = self.w.size
            elif len(final_shape) == self.pdc.shape[0] - 1:
                # append the shape of the polariation-detection axis
                final_shape.append(self.w.size)
            else:
                raise Exception('Cannot automatically determine final signal shape')
                
        else:
            initial_shape = (len(all_delay_combinations),)

        return initial_shape, final_shape
        
    def calculate_diagrams_all_delays(self,diagrams):
        t0 = time.time()

        all_delay_combinations = list(itertools.product(*self.all_pulse_delays))
        initial_shape, final_shape = self.get_initial_and_final_signal_shapes(all_delay_combinations)
        signal = np.zeros(initial_shape,dtype='complex')

        counter = 0
        for delays in all_delay_combinations:
            arrival_times = [0]
            for delay in delays:
                arrival_times.append(arrival_times[-1]+delay)

            if self.detection_type == 'polarization' or self.detection_type == 'complex_polarization':
                signal[counter,:] = self.calculate_diagrams(diagrams,arrival_times)
            else:
                signal[counter] = self.calculate_diagrams(diagrams,arrival_times)
            counter += 1

        self.signal = signal.reshape(final_shape)
        if not self.return_complex_signal:
            self.signal = np.real(self.signal)
        self.calculation_time = time.time() - t0
        
        return self.signal

    def set_pulse_times(self,arrival_times):
        if hasattr(self,'pulse_times'):
            old_pulse_times = self.pulse_times
            for i in range(len(old_pulse_times)):
                if old_pulse_times[i] != arrival_times[i]:
                    self.remove_composite_calculations_by_pulse_number(i)
                    self.remove_calculations_by_pulse_number(i)
        else:
            self.reset()
        
        self.pulse_times = arrival_times
        
    def calculate_diagrams(self,diagram_instructions,arrival_times):
        self.set_pulse_times(arrival_times)
        
        self.current_instructions = diagram_instructions
        instructions = diagram_instructions[0]
        signal = self.execute_diagram(instructions)
        rho = self.execute_diagram(instructions)
        signal = self.rho_to_signal(rho)
        for instructions in diagram_instructions[1:]:
            rho = self.execute_diagram(instructions)
            signal += self.rho_to_signal(rho)
        return signal

    def calculate_signal_all_delays(self,*,composite_diagrams=False):
        
        def calculate_signal(t):
            return self.calculate_signal(t,composite_diagrams=composite_diagrams)
        
        t0 = time.time()

        all_delay_combinations = list(itertools.product(*self.all_pulse_delays))

        initial_shape, final_shape = self.get_initial_and_final_signal_shapes(all_delay_combinations)

        signal = np.zeros(initial_shape,dtype='complex')

        if not self.return_complex_signal:
            signal = np.real(signal)

        signal_dict = {key:signal.copy() for key in self.signal_pdcs}

        counter = 0
        for delays in all_delay_combinations:
            arrival_times = [0]
            for delay in delays:
                arrival_times.append(arrival_times[-1]+delay)

            if self.detection_type == 'polarization' or self.detection_type == 'complex_polarization':
                single_signal_dict = calculate_signal(arrival_times)
                for key in self.signal_pdcs:
                    signal_dict[key][counter,:] = single_signal_dict[key]
            else:
                single_signal_dict = calculate_signal(arrival_times)
                for key in self.signal_pdcs:
                    signal_dict[key][counter] = single_signal_dict[key]
            counter += 1

        for key in self.signal_pdcs:
            signal_dict[key] = signal_dict[key].reshape(final_shape)
            
        self.calculation_time = time.time() - t0
        
        self.signal_dict = signal_dict
        if len(self.signal_pdcs) == 1:
            key = self.signal_pdcs[0]
            self.signal = signal_dict[key]
            return self.signal
        else:
            return self.signal_dict

    def set_t_general(self,optical_dephasing_rate,*,dt='auto'):
        """Sets the time grid upon which all frequency-detected signals will
            be calculated on
"""
        max_pos_t = int(self.gamma_res/optical_dephasing_rate)
        max_efield_t = max([np.max(u) for u in self.efield_times]) * 1.05
        max_pos_t = max(max_pos_t,max_efield_t)
        if dt == 'auto':
            dt = self.dts[-1] # signal detection bandwidth determined by local oscillator
        n = int(max_pos_t/dt)
        self.t = np.arange(-n,n+1,1)*dt
        self.w = fftshift(fftfreq(self.t.size,d=dt)*2*np.pi)
        
    def polarization_detection_rho_to_signal(self,rho):
        t0 = time.time()
        p_of_t = self.dipole_expectation(rho,ket_flag=True)
        self.rho_to_polarization_time += time.time() - t0
        return self.polarization_to_signal(p_of_t,local_oscillator_number=-1)

    def integrated_polarization_detection_rho_to_signal(self,rho):
        t0 = time.time()
        p = self.integrated_dipole_expectation(rho,ket_flag=True)
        self.rho_to_polarization_time += time.time() - t0
        return self.integrated_polarization_to_signal(p,local_oscillator_number=-1)

    def fluorescence_detection_rho_to_signal(self,rho):
        t = np.array([rho.t0 + 1E-2])
        signal = self.fl_exp_val(t,rho)
        return signal
    
    def set_local_oscillator_phase(self,phase):
        self.efields[-1] = np.exp(1j*phase) * self.local_oscillator

    def calculate_signal(self,arrival_times,*,composite_diagrams=False):
        self.set_pulse_times(arrival_times)

        signal_dict = {}
        
        if composite_diagrams:
            self.execute_composite_diagrams()
            for key in self.signal_pdcs:
                signal = self.key_to_signal(key)
                if not self.return_complex_signal:
                    signal = np.real(signal)
                signal_dict[key] = signal

        else:
            t0 = time.time()
            self.set_current_diagram_instructions(arrival_times)
            t1 = time.time()

            for key in self.signal_pdcs:
                diagrams = self.current_instructions[key]
                if len(diagrams) == 0:
                    signal = 0
                else:
                    instructions = diagrams[0]
                    rho = self.execute_diagram(instructions)
                    signal = self.rho_to_signal(rho)
                    for instructions in diagrams[1:]:
                        rho = self.execute_diagram(instructions)
                        signal += self.rho_to_signal(rho)
                    if not self.return_complex_signal:
                        signal = np.real(signal)

                signal_dict[key] = signal

            t2 = time.time()
            self.automation_time += t1-t0
            self.diagram_to_signal_time += t2-t1

        return signal_dict

    def set_undersample_factor(self,frequency_resolution):
        """dt is set by the pulse. However, the system dynamics may not require such a 
            small dt.  Therefore, this allows the user to set a requested frequency
            resolution for any spectrally resolved signals.  This assumes that the 
            undersampling is defined with respect to the last pulse, which is assumed
            to be the local oscillator"""
        # f = pi/dt
        dt = np.pi/frequency_resolution
        u = int(np.floor(dt/self.dts[-1]))
        self.undersample_factor = max(u,1)
        
    def dipole_expectation(self,rho_in,*,ket_flag=True):
        """Computes the expectation value of the dipole operator"""
        t0 = time.time()

        pulse_number = -1

        # if rho_in.pulse_number == None:
        #     old_pulse_time = 0
        # else:
        #     old_pulse_time = self.pulse_times[rho_in.pulse_number]
        pulse_time = self.pulse_times[pulse_number]

        efield_t = self.efield_times[pulse_number] + pulse_time

        if ket_flag:
            center = - self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]

        # The signal is zero before the final pulse arrives, and persists
        # until it decays. Therefore we avoid taking the sum at times
        # where the signal is zero.
        t = self.t + pulse_time

        pulse_start_ind = np.argmin(np.abs(t-efield_t[0]))
        if efield_t[0] < t[pulse_start_ind]:
            pulse_start_ind -= 1
        pulse_end_ind = np.argmin(np.abs(t-efield_t[-1]))
        
        if efield_t[-1] > t[pulse_end_ind]:
            pulse_end_ind += 1

        t0a = time.time()
        t_slice = slice(pulse_start_ind,None,None)

        t = self.t[t_slice] + pulse_time
        
        u = self.undersample_factor
        t1_slice = slice(pulse_start_ind,pulse_end_ind,None)
        t2_slice = slice(pulse_end_ind,None,u)

        t1 = self.t[t1_slice] + pulse_time
        t2 = self.t[t2_slice] + pulse_time

        t_u = np.hstack((t1,t2))
        

        t0b = time.time()
        if t1.size == 0:
            exp_val1 = np.array([])
        else:
            exp_val1 = self.mu_exp_val(t1,rho_in,pulse_number,mu_type='H',
                                       ket_flag = ket_flag, up_flag = False)
        # if self.method == 'chebyshev':
        #     dom = np.array([t2[0],t2[-1]])
        #     halfwidth = (dom[1] - dom[0])/2
        #     midpoint = (dom[1] + dom[0])/2
        #     M = 61
        #     tcheb = npch.chebpts1(M) * halfwidth + midpoint
        #     exp_val_ch=self.mu_exp_val(tcheb,rho_in,pulse_number,mu_type='H',
        #                                ket_flag = ket_flag, up_flag = False)
        #     exp_val_rot = exp_val_ch * np.exp(-1j * center * tcheb)
        #     exp_val_poly = ChebPoly(tcheb,exp_val_rot,dom=dom)
        #     exp_val2 = exp_val_poly(t2)[0,:]
        #     exp_val2 = exp_val2 * np.exp(+1j * center * t2)
        # else:
        exp_val2 = self.mu_exp_val(t2,rho_in,pulse_number,mu_type='H',
                                   ket_flag = ket_flag, up_flag = False)
        
        tb = time.time()
        self.expectation_time += tb-t0b

        exp_val_u = np.hstack((exp_val1,exp_val2))
        
        # rotate away carrier frequency
        exp_val_u  = exp_val_u * np.exp(-1j * center * t_u)

        ta = time.time()
        
        self.slicing_time += ta - t0a - (tb - t0b)

        t0 = time.time()

        # Interpolate expectation value back onto the full t-grid
        if u != 1:
            # Often must extrapolate the final point
            exp_val_interp = scipy.interpolate.interp1d(t_u,exp_val_u,kind='cubic',fill_value='extrapolate')
            exp_val = exp_val_interp(t)
        else:
            exp_val = exp_val_u

        # Initialize return array with zeros
        ret_val = np.zeros(self.t.size,dtype='complex')
        
        # set non-zero values using t_slice
        ret_val[pulse_start_ind:] = exp_val

        tb = time.time()
        self.interpolation_time += tb-t0
        
        return ret_val

    def integrated_dipole_expectation(self,rho_in,*,ket_flag=True):
        """Computes the expectation value of the dipole operator"""
        
        pulse_number = -1
        
        pulse_time = self.pulse_times[pulse_number]
        t = pulse_time + self.efield_times[pulse_number]

        if ket_flag:
            center = - self.centers[pulse_number]
        else:
            center = self.centers[pulse_number]

        t0 = time.time()
        exp_val = self.mu_exp_val(t,rho_in,pulse_number,mu_type='H',
                                  ket_flag = ket_flag, up_flag = False)

        # rotate away carrier frequency
        exp_val  = exp_val * np.exp(-1j * center * t)
        
        tb = time.time()
        self.expectation_time += tb-t0

        return exp_val

    def get_local_oscillator(self):
        local_oscillator_number = -1
        efield_t = self.efield_times[local_oscillator_number]
        efield = self.efields[local_oscillator_number]

        if efield_t.size == 1:
            # Impulsive limit: delta in time is flat in frequency
            efield_ft = np.ones(self.w.size)*efield
            return efield_ft

        if self.method == 'chebyshev':
            dom = self.doms[local_oscillator_number]
            chp = ChebPoly(efield_t,efield[np.newaxis,:],dom = dom)
            full_efield = chp(self.t)[0,:]
            dt = self.t[1] - self.t[0]
            efield_ft = fftshift(ifft(ifftshift(full_efield)))*full_efield.size * dt
            return efield_ft
        
        e_dt = efield_t[1] - efield_t[0]
        dt = self.t[1] - self.t[0]

        if (not np.isclose(e_dt,dt) and e_dt > dt):
            raise Exception("""Local oscillator dt is too large for desired 
                signal detection bandwidth.  You must either use method
                set_t with a larger dt, or supply local oscillator with 
                smaller value of dt""")
        elif (np.isclose(e_dt,dt) and efield_t[-1] >= self.t[-1]):
            full_efield = np.zeros(self.t.size,dtype='complex')

            # the local oscillator sets the "zero" on the clock
            pulse_time_ind = np.argmin(np.abs(self.t))

            pulse_start_ind = pulse_time_ind - efield_t.size//2
            pulse_end_ind = pulse_time_ind + efield_t.size//2 + efield_t.size%2

            t_slice = slice(pulse_start_ind, pulse_end_ind,None)
            
            full_efield[t_slice] = efield
            efield_ft = fftshift(ifft(ifftshift(full_efield)))*full_efield.size * dt
        elif efield_t[-1] > self.t[-1]:
            f = sinterp1d(efield_t,efield,fill_value = (0,0),bounds_error=False,
                          kind='linear')
            full_efield = f(self.t)
            efield_ft = fftshift(ifft(ifftshift(full_efield)))*full_efield.size * dt
        else:
            efield_ft = fftshift(ifft(ifftshift(efield))) * efield.size * e_dt
            efield_w = fftshift(fftfreq(efield_t.size,d=e_dt)) * 2 * np.pi
            fill_value = (efield_ft[0],efield_ft[-1])
            f = sinterp1d(efield_w,efield_ft,fill_value = fill_value,
                          bounds_error=False,kind='quadratic')
            efield_ft = f(self.w)

        return efield_ft
    
    def polarization_to_signal(self,P_of_t_in,*,
                                local_oscillator_number = -1,undersample_factor = 1):
        """This function generates a frequency-resolved signal from a polarization field
           local_oscillator_number - usually the local oscillator will be the last pulse 
                                     in the list self.efields"""
        undersample_slice = slice(None,None,undersample_factor)
        P_of_t = P_of_t_in[undersample_slice]#.copy()
        t = self.t[undersample_slice]
        dt = t[1] - t[0]
        if self.add_lorentzian_linewidth:
            P_of_t = P_of_t * np.exp(-self.gamma*np.abs(t))
        pulse_time = self.pulse_times[local_oscillator_number]

        efield = self.get_local_oscillator()

        halfway = self.w.size//2
        pm = self.w.size//(2*undersample_factor)
        efield_min_ind = halfway - pm
        efield_max_ind = halfway + pm + self.w.size%2
        efield = efield[efield_min_ind:efield_max_ind]

        P_of_w = fftshift(ifft(ifftshift(P_of_t)))*P_of_t.size*dt/np.pi

        signal = P_of_w * np.conjugate(efield)
        if not self.return_complex_signal:
            return np.imag(signal)
        else:
            return -1j*signal

    def integrated_polarization_to_signal(self,P,*,
                                local_oscillator_number = -1):
        """This function generates a frequency-resolved signal from a polarization field
           local_oscillator_number - usually the local oscillator will be the last pulse 
                                     in the list self.efields"""
        efield_t = self.efield_times[local_oscillator_number]

        efield = self.efields[local_oscillator_number]

        if len(P) == 1:
            signal = P * np.conjugate(efield)

        else:
            signal = np.trapz(P * np.conjugate(efield),x=efield_t)
            
        if not self.return_complex_signal:
            return np.imag(signal)
        else:
            return 1j*signal

    def add_gaussian_linewidth(self,sigma):
        try:
            old_signal_dict = self.old_signal_dict
        except AttributeError:
            self.old_signal_dict = copy.deepcopy(self.signal_dict)
            old_signal_dict = self.old_signal_dict

        for pdc in self.signal_pdcs:
            signal = self.old_signal_dict[pdc]

            if len(signal.shape) == 3:

                sig_tau_t = fftshift(fft(ifftshift(signal,axes=(-1)),axis=-1),axes=(-1))
                sig_tau_t = sig_tau_t * (np.exp(-self.t**2/(2*sigma**2))[np.newaxis,np.newaxis,:]
                                         *np.exp(-self.t21_array**2/(2*sigma**2))[:,np.newaxis,np.newaxis])
                sig_tau_w = fftshift(ifft(ifftshift(sig_tau_t,axes=(-1)),axis=-1),axes=(-1))

                new_signal = sig_tau_w

            elif len(signal.shape) == 2:
                sig_t = fftshift(fft(ifftshift(signal,axes=(-1)),axis=-1),axes=(-1))
                sig_t = sig_t * (np.exp(-self.t**2/(2*sigma**2))[np.newaxis,:])
                sig_w = fftshift(ifft(ifftshift(sig_t,axes=(-1)),axis=-1),axes=(-1))

                new_signal = sig_w

            self.signal_dict[pdc] = new_signal

    def save(self,file_name,pulse_delay_names = [],*,use_base_path=True,makedir=True):
        if use_base_path:
            file_name = os.path.join(self.base_path,file_name)
        if makedir:
            folder = os.path.split(file_name)[0]
            os.makedirs(folder,exist_ok=True)
        if len(pulse_delay_names) == 0:
            pulse_delay_names = ['t' + str(i) for i in range(len(self.all_pulse_delays))]
        try:
            save_dict = {'signal':self.signal}
        except AttributeError:
            save_dict = {}
        if 'signal_dict' in self.__dict__.keys():
            for key in self.signal_dict.keys():
                save_dict[str(key)] = self.signal_dict[key]

        for name,delays in zip(pulse_delay_names,self.all_pulse_delays):
            save_dict[name] = delays
        save_dict['efields'] = self.efields
        save_dict['efield_times'] = self.efield_times
        save_dict['centers'] = self.centers
        save_dict['pdc'] = self.pdc
        save_dict['polarization_sequence'] = self.polarization_sequence
        if self.detection_type == 'polarization' or self.detection_type == 'complex_polarization':
            save_dict['wt'] = self.w
        save_dict['signal_calculation_time'] = self.calculation_time
        np.savez(file_name,**save_dict)

    def load(self,file_name,pulse_delay_names=[],*,use_base_path=True):
        if use_base_path:
            file_name = os.path.join(self.base_path,file_name)
        arch = np.load(file_name)
        print(list(arch.keys()))
        self.all_pulse_delays = []
        if len(pulse_delay_names) == 0:
            for key in arch.keys():
                if key[0] == 't':
                    pulse_delay_names.append(key)
        if len(pulse_delay_names) == 0:
            raise Exception('Pulse delay names not in standard format, must specify manually')
        for name in pulse_delay_names:
            self.all_pulse_delays.append(arch[name])
        if self.detection_type == 'polarization' or self.detection_type == 'complex_polarization':
            try:
                self.w = arch['wt']
                dw = self.w[1] - self.w[0]
                self.t = fftshift(fftfreq(self.w.size,d=dw)*2*np.pi)
            except KeyError:
                print('No detection frequency found')
        self.signal_dict = {}
        self.signal_pdcs = []
        for key in arch.keys():
            if key[0:2] == '((':
                self.signal_dict[eval(key)] = arch[key]
                self.signal_pdcs.append(eval(key))
        self.calculation_time = arch['signal_calculation_time']
        try:
            self.signal = arch['signal']
        except KeyError:
            pass
        try:
            self.efields = arch['efields']
            self.efield_times = arch['efield_times']
            self.centers = arch['centers']
            self.pdc = arch['pdc']
            self.polarization_sequence = arch['polarization_sequence']
        except KeyError:
            pass

class CalculateSignalsOpen(CalculateSignals):

    def set_t(self,optical_dephasing_rate,*,dt='auto'):
        """Sets the time grid upon which all frequency-detected signals will
            be calculated on
"""
        if optical_dephasing_rate == 'auto':
            try:
                eigvals = self.eigenvalues['0,1']
            except KeyError:
                eigvals = self.eigenvalues['all_manifolds']
            optical_dephasing_rate = -np.min(np.real(eigvals))
        self.gamma = optical_dephasing_rate
        self.set_t_general(optical_dephasing_rate,dt=dt)

    def key_to_signal(self,signal_key):
        rho = self.composite_rhos[signal_key]
        return self.rho_to_signal(rho)

class CalculateSignalsClosed(CalculateSignals):

    def set_t(self,optical_dephasing_rate,*,dt='auto',set_gamma=True):
        """Sets the time grid upon which all frequency-detected signals will
            be calculated on
"""
        if set_gamma:
            self.gamma = optical_dephasing_rate
        else:
            self.gamma = 0
        self.set_t_general(optical_dephasing_rate,dt=dt)

    def key_to_signal(self,signal_key):
        calculation_keys = []
        signal_pdc = self.pdc_tup_to_arr(signal_key)
        if self.detection_type == 'integrated_polarization':
            signal = np.zeros(1,dtype='complex')
        else:
            signal = np.zeros(self.w.size,dtype='complex')
        for key in self.composite_psis.keys():
            psi = self.composite_psis[key]
            ket_pdc_a = psi.pdc
            remaining_pdc_a = signal_pdc - ket_pdc_a
            bra_pdc_a = remaining_pdc_a[:,::-1]
            bra_key_a = self.pdc_arr_to_tup(bra_pdc_a)
            if bra_key_a in self.composite_psis.keys():
                calculation_key = (bra_key_a,key)
                if calculation_key in calculation_keys:
                    pass
                else:
                    calculation_keys.append(calculation_key)
                    psi_bra_a = self.composite_psis[bra_key_a]
                    rho = (psi_bra_a,psi)
                    signal += self.rho_to_signal(rho)

            bra_pdc_b = psi.pdc[:,::-1]
            ket_pdc_b = signal_pdc - bra_pdc_b
            ket_key_b = self.pdc_arr_to_tup(ket_pdc_b)
            if ket_key_b in self.composite_psis.keys():
                calculation_key = (key,ket_key_b)
                if calculation_key in calculation_keys:
                    pass
                else:
                    calculation_keys.append(calculation_key)
                    psi_ket_b = self.composite_psis[ket_key_b]
                    rho = (psi,psi_ket_b)
                    signal += self.rho_to_signal(rho)
        return signal

    
class UF2DensityMatrices(UF2OpenEngine,CalculateSignalsOpen):
    def __init__(self,file_path,*,detection_type='polarization',
                 conserve_memory=False):
        UF2OpenEngine.__init__(self,file_path,detection_type=detection_type,
                           conserve_memory=conserve_memory)
        CalculateSignalsOpen.__init__(self,detection_type=detection_type)

class RKDensityMatrices(RKOpenEngine,CalculateSignalsOpen):
    def __init__(self,file_path,*,detection_type='polarization',
                 conserve_memory=False):
        RKOpenEngine.__init__(self,file_path,detection_type=detection_type,
                           conserve_memory=conserve_memory)
        CalculateSignalsOpen.__init__(self,detection_type=detection_type)

class UF2Wavefunctions(UF2ClosedEngine,CalculateSignalsClosed):
    def __init__(self,file_path,*,detection_type='polarization',
                conserve_memory=None):
        UF2ClosedEngine.__init__(self,file_path,detection_type=detection_type)
        CalculateSignalsClosed.__init__(self,detection_type=detection_type)

        self.add_lorentzian_linewidth = True

class RKWavefunctions(RKClosedEngine,CalculateSignalsClosed):
    def __init__(self,file_path,*,detection_type='polarization',
                conserve_memory=None):
        RKClosedEngine.__init__(self,file_path,detection_type=detection_type)
        CalculateSignalsClosed.__init__(self,detection_type=detection_type)

        self.add_lorentzian_linewidth = True


class SpectroscopyBase:
    def __init__(self,file_name,*,engine_name = 'UF2',include_linear=False,
                 detection_type = 'polarization',conserve_memory=False):
        if file_name[-1] == '/':
            file_name = file_name[:-1]
        file_end = os.path.split(file_name)[-1]
        if file_end == 'closed':
            if engine_name == 'UF2':
                engine = UF2Wavefunctions
            elif engine_name == 'RK':
                engine = RKWavefunctions
                
        elif file_end == 'open':
            if engine_name == 'UF2':
                engine = UF2DensityMatrices
            elif engine_name == 'RK':
                engine = RKDensityMatrices
        self.calc = engine(file_name,conserve_memory = conserve_memory,
                           detection_type=detection_type)
        self.engine = self.calc
        self.save_name = 'signals'

    def set_phase_discrimination(self,pdc):
        self.pdc = pdc

    def set_efields(self,efield_times,efields,centers):
        self.engine.set_efields(efield_times,efields,centers,self.pdc)
        self.engine.set_polarization_sequence(['x']*len(self.pdc))

    def set_t(self,gamma,*,dt='auto'):
        self.engine.set_t(gamma,dt=dt)

    def set_identical_gaussians(self,sigma,center,Delta = 10, M = 41):
        self.engine.set_identical_gaussians(sigma,center,self.pdc,
                                          Delta = Delta, M = M)
        self.engine.set_polarization_sequence(['x']*len(self.pdc))
        self.engine.centers = [center]*len(self.pdc)

    def set_impulsive_pulses(self,center):
        self.engine.set_impulsive_pulses(self.pdc)
        self.engine.set_polarization_sequence(['x']*len(self.pdc))
        self.engine.centers = [center]*len(self.pdc)
            
    def set_highest_order(self,highest_order):
        self.engine.include_higher_orders((highest_order-1,1))
    
    def set_pulse_delays(self,pulse_delays):
        self.engine.set_pulse_delays(pulse_delays)

    def calculate_signal_all_delays(self,*,composite_diagrams=True):
        self.engine.calculate_signal_all_delays(composite_diagrams=composite_diagrams)
        self.engine.save(self.save_name)

    def get_signal(self,lam_list,*,max_order = np.inf):
        """lam is the perturbative parameter
        Args:
            lam (float) : unit-less electric field amplitude scaling factor
"""
        random_key = self.engine.signal_pdcs[0]
        signal = np.zeros(self.engine.signal_dict[random_key].shape,dtype='complex')
        for key in self.engine.signal_pdcs:
            pulse_orders = self.engine.pdc_tup_to_arr(key).sum(axis=1)
            if np.sum(pulse_orders) <= max_order:
                lam_powers = [l**p_ord for l,p_ord in zip(lam_list,pulse_orders)]
                perturbative_parameter = np.prod(lam_powers)
                signal += self.engine.signal_dict[key] * perturbative_parameter

        return signal

    def get_signal_order(self,order):
        random_key = self.engine.signal_pdcs[0]
        signal = np.zeros(self.engine.signal_dict[random_key].shape,dtype='complex')
        for key in self.engine.signal_pdcs:
            if self.engine.pdc_tup_to_arr(key).sum() == order:
                signal += self.engine.signal_dict[key]

        return signal



import unittest
import numpy as np
import os
import yaml

import ufss

def setup_params():
    """This function creates the file 'simple_params.yaml' inside the 
        specified folder. The Hamiltonian/Liouvillian Generator (HLG) part of
        ufss needs 'simple_params.yaml' to run.
"""
    truncation_size = 1 # irrelevant since no vibrations are included
    folder = os.path.join('fixtures','vibrationless_dimer')
    os.makedirs(folder,exist_ok=True)

    ### Define Hamiltonian
    site_energies = [0,0.75]
    site_couplings = [0.3307]

    vibs = []

    ### Define electronic dipoles
    # dipoles are input as a list of cartesian vectors [mu_x, mu_y, mu_z]
    mu = [[1,0,0],[1,0,0]]

    # Use a closed system for this calculation, which is acheived by
    # using 'coupling':0. This creates a closed folder and an open folder,
    # which should produce identical results, if all is working correctly
    # all other bath paramters are irrelevant in this case
    site_bath = {'cutoff_frequency':1,# in units of omega_0
                 'coupling':0,# system-bath coupling in units of omega_0
                 'temperature':1,# kT in units of omega_0
                 'cutoff_function':'lorentz-drude',
                 'spectrum_type':'ohmic'}

    Redfield_bath = {'secular':False, 
                     'site_bath':site_bath,
                     'vibration_bath':site_bath}

    params = {
        'site_energies':site_energies,
        'site_couplings':site_couplings,
        'dipoles':mu,
        'truncation_size':truncation_size,
        'vibrations':vibs,
        'bath':Redfield_bath}

    with open(os.path.join(folder,'simple_params.yaml'),'w+') as file_stream:
        yaml.dump(params,file_stream)

    ufss.HLG.run(folder)
    return None

def calculate_linear_absorption_with_UF2(closed_system=True):
    """Calculate the linear absorption spectrum using either the closed- or
        open-systems methods of UF2.
"""
    folder = os.path.join('fixtures','vibrationless_dimer')

    if closed_system == True:
        la = ufss.Wavepackets(os.path.join(folder,'closed'),
                              detection_type='polarization')
    else:
        la = ufss.DensityMatrices(os.path.join(folder,'open'),
                              detection_type='polarization')

    # defining the optical pulses in the RWA
    M = 41 # number of points used to resolve optical pulses
    Delta = 10 # pulse interval
    t = np.linspace(-Delta/2,Delta/2,num=M)
    dt = t[1] - t[0]

    sigma = 1
    ef = np.zeros(t.size,dtype='float')
    ef[ef.size//2] = 1/dt

    la.set_polarization_sequence(['x'])

    la.set_efields([t],[ef],[0],[(1,0)])

    gamma_dephasing = 0.2/sigma
    la.gamma_res = 20
    if closed_system == True:
        la.set_t(gamma_dephasing,set_gamma=False)
    else:
        la.set_t(gamma_dephasing)
    la.set_pulse_delays([])

    sig = la.calculate_signal_all_delays()

    return la.w, sig

def calculate_peak_ratio(w,sig):
    """Calculate the integrated peak ratios of a 1D signal, assuming that
        the peaks are centered at 0 and 1 omega0, respectively.
        Args:
            w (np.array) : 1D array of frequencies
            sig (np.array) : 1D array of the signal
        Returns:
            ratio of the integrated peak areas
"""
    peak0_inds = np.where((w>-0.5) & (w<0.5))[0]
    peak0_area = np.trapz(sig[peak0_inds],x=w[peak0_inds])

    peak1_inds = np.where((w>0.5) & (w<1.5))[0]
    peak1_area = np.trapz(sig[peak1_inds],x=w[peak1_inds])

    return peak0_area/peak1_area
    

class test_against_squared_dipoles(unittest.TestCase):
    """Check to make sure that the calculated linear absorption spectrum
        for a vibration-less dimer gives two peaks at 0 and 1 omega_0, and 
        that the integrated peak ratios match the ratio mu_0^2/mu_1^2. Note
        that this does not really check the HLG, since we are not asserting 
        what the values of mu_0 and mu_1 are, but are simply loading them
        from the file mu.npz
"""
    def test_open(self):
        # Use the HLG to set up the system
        setup_params()
        # Calculate the linear absorption spectrum
        w, sig = calculate_linear_absorption_with_UF2(closed_system=False)
        # Load the dipole operator that is fed to UF2
        mu_path = os.path.join('fixtures','vibrationless_dimer',
                               'closed','mu.npz')
        mu = np.load(mu_path)
        expected_peak_ratio = mu['0_to_1'][0,0,0]**2/mu['0_to_1'][1,0,0]**2
        actual_peak_ratio = calculate_peak_ratio(w,sig)
        abs_err = np.abs(expected_peak_ratio-actual_peak_ratio)
        rel_err = np.abs(abs_err/actual_peak_ratio)
        # I have found that the relative error is roughly 0.000497
        # I do not expect perfect agreement, since I am using a discrete
        # Fourier transform. Increasing the frequency resolution of the
        # signal should diminish the error
        self.assertTrue(rel_err<0.0005)

    def test_closed(self):
        # Use the HLG to set up the system
        setup_params()
        # Calculate the linear absorption spectrum
        w, sig = calculate_linear_absorption_with_UF2(closed_system=True)
        self.open_sig = sig.copy()
        # Load the dipole operator that is fed to UF2
        mu_path = os.path.join('fixtures','vibrationless_dimer',
                               'closed','mu.npz')
        mu = np.load(mu_path)
        expected_peak_ratio = mu['0_to_1'][0,0,0]**2/mu['0_to_1'][1,0,0]**2
        actual_peak_ratio = calculate_peak_ratio(w,sig)
        abs_err = np.abs(expected_peak_ratio-actual_peak_ratio)
        rel_err = np.abs(abs_err/actual_peak_ratio)
        # I have found that the relative error is roughly 0.000497
        # I do not expect perfect agreement, since I am using a discrete
        # Fourier transform. Increasing the frequency resolution of the
        # signal should diminish the error
        self.assertTrue(rel_err<0.0005)

    def test_open_vs_closed(self):
        """For linear signals, the open and closed case should be identical
            in the case of 0 coupling to the bath. Even when the bath 
            coupling is 0, this is not quite the case for nonlinear signals.
"""
        w, open_sig = calculate_linear_absorption_with_UF2(closed_system=False)
        w, closed_sig = calculate_linear_absorption_with_UF2(closed_system=True)
        self.assertTrue(np.allclose(closed_sig,open_sig))
        
if __name__ == '__main__':
    unittest.main()
np.load(analytical_signal.npy)



import numpy as np
import scipy.sparse as sp

class LindbladConstructor:

    @staticmethod
    def make_Lindblad_instructions(gamma,O):
        """O must be square
"""
        II = np.eye(O.shape[0])
        Od = np.conjugate(O.T)
        leftright = gamma * (-np.dot(Od,O)/2)
        return [(gamma*O,Od),(leftright,II),(II,leftright)]

    @staticmethod
    def make_Lindblad_instructions2(gamma,Oket,Obra):
        IIket = np.eye(Oket.shape[0])
        IIbra = np.eye(Obra.shape[0])
        Oketd = np.conjugate(Oket.T)
        Obrad = np.conjugate(Obra.T)
        left = gamma * (-np.dot(Oketd,Oket)/2)
        right = gamma * (-np.dot(Obrad,Obra)/2)
        return [(gamma*Oket,Obrad),(left,IIbra),(IIket,right)]

    @staticmethod
    def make_Lindblad_instructions2_Obra0(gamma,Oket,Obra):
        IIbra = np.eye(Obra.shape[0])
        Oketd = np.conjugate(Oket.T)
        left = gamma * (-np.dot(Oketd,Oket)/2)
        return [(left,IIbra)]

    @staticmethod
    def make_Lindblad_instructions2_Oket0(gamma,Oket,Obra):
        IIket = np.eye(Oket.shape[0])
        Obrad = np.conjugate(Obra.T)
        right = gamma * (-np.dot(Obrad,Obra)/2)
        return [(IIket,right)]

class LiouvillianConstructor(LindbladConstructor):

    @staticmethod
    def make_commutator_instructions(O):
        """O must be square
"""
        II = np.eye(O.shape[0])
        return [(O,II),(II,-O)]

    @staticmethod
    def make_commutator_instructions2(Oket,Obra):
        """
"""
        IIket = np.eye(Oket.shape[0])
        IIbra = np.eye(Obra.shape[0])
        return [(Oket,IIbra),(IIket,-Obra)]

    @staticmethod
    def make_Liouvillian(instruction_list,*,sparse=False):
        if sparse:
            kron = sp.kron
        else:
            kron = np.kron
        left, right = instruction_list[0]
        L = kron(left,right.T)
        for left,right in instruction_list[1:]:
            L = L + kron(left,right.T)
        return L
